import torch
from torch import sin, cos
from torch import autograd
from neurodiffeq.neurodiffeq import safe_diff as diff


def grad(u, *xs):
    r"""Gradient of tensor u with respect to a tuple of tensors xs.
    Given :math:`u` and :math:`x_1`, ..., :math:`x_n`, the function returns
    :math:`\frac{\partial u}{\partial x_1}`, ..., :math:`\frac{\partial u}{\partial x_n}`

    :param u: The :math:`u` described above.
    :type u: `torch.Tensor`
    :param xs: The sequence of :math:`x_i` described above.
    :type xs: `torch.Tensor`
    :return: A tuple of :math:`\frac{\partial u}{\partial x_1}`, ..., :math:`\frac{\partial u}{\partial x_n}`
    :rtype: List[`torch.Tensor`]
    """
    grads = []
    for x, g in zip(xs, autograd.grad(u, xs, grad_outputs=torch.ones_like(u), create_graph=True, allow_unused=True)):
        if g is None:
            grads.append(torch.zeros_like(x, requires_grad=True))
        else:
            grads.append(g.requires_grad_(True))
    return grads


def spherical_curl(u_r, u_theta, u_phi, r, theta, phi):
    r"""Derives and evaluates the spherical curl of a spherical vector field :math:`u`.

    :param u_r: The :math:`r`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_r: `torch.Tensor`
    :param u_theta: The :math:`\theta`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_theta: `torch.Tensor`
    :param u_phi: The :math:`\phi`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_phi: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The :math:`r`, :math:`\theta`, and :math:`\phi` components of the curl, each with shape (n_samples, 1).
    :rtype: tuple[`torch.Tensor`]
    """

    d_r = lambda u: diff(u, r)
    d_theta = lambda u: diff(u, theta)
    d_phi = lambda u: diff(u, phi)

    curl_r = (d_theta(u_phi * sin(theta)) - d_phi(u_theta)) / (r * sin(theta))
    curl_theta = (d_phi(u_r) / sin(theta) - d_r(u_phi * r)) / r
    curl_phi = (d_r(u_theta * r) - d_theta(u_r)) / r

    return curl_r, curl_theta, curl_phi


def spherical_grad(u, r, theta, phi):
    r"""Derives and evaluates the spherical gradient of a spherical scalar field :math:`u`.

    :param u: A scalar field :math:`u`, must have shape (n_samples, 1).
    :type u: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The :math:`r`, :math:`\theta`, and :math:`\phi` components of the gradient, each with shape (n_samples, 1).
    :rtype: tuple[`torch.Tensor`]
    """
    grad_r = diff(u, r)
    grad_theta = diff(u, theta) / r
    grad_phi = diff(u, phi) / (r * sin(theta))
    return grad_r, grad_theta, grad_phi


def spherical_div(u_r, u_theta, u_phi, r, theta, phi):
    r"""Derives and evaluates the spherical divergence of a spherical vector field :math:`u`.

    :param u_r: The :math:`r`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_r: `torch.Tensor`
    :param u_theta: The :math:`\theta`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_theta: `torch.Tensor`
    :param u_phi: The :math:`\phi`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_phi: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The divergence evaluated at :math:`(r, \theta, \phi)`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    div_r = diff(u_r * r ** 2, r) / r ** 2
    div_theta = diff(u_theta * sin(theta), theta) / (r * sin(theta))
    div_phi = diff(u_phi, phi) / (r * sin(theta))
    return div_r + div_theta + div_phi


def spherical_laplacian(u, r, theta, phi):
    r"""Derives and evaluates the spherical laplacian of a spherical scalar field :math:`u`.

    :param u: A scalar field :math:`u`, must have shape (n_samples, 1).
    :type u: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The laplacian evaluated at :math:`(r, \theta, \phi)`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    d_r = lambda u: diff(u, r)
    d_theta = lambda u: diff(u, theta)
    d_phi = lambda u: diff(u, phi)

    lap_r = d_r(d_r(u) * r ** 2) / r ** 2
    lap_theta = d_theta(d_theta(u) * sin(theta)) / (r ** 2 * sin(theta))
    lap_phi = d_phi(d_phi(u)) / (r ** 2 * sin(theta) ** 2)
    return lap_r + lap_theta + lap_phi


def spherical_vector_laplacian(u_r, u_theta, u_phi, r, theta, phi):
    r"""Derives and evaluates the spherical laplacian of a spherical vector field :math:`u`.

    :param u_r: The :math:`r`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_r: `torch.Tensor`
    :param u_theta: The :math:`\theta`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_theta: `torch.Tensor`
    :param u_phi: The :math:`\phi`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_phi: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The laplacian evaluated at :math:`(r, \theta, \phi)`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    d_theta = lambda u: diff(u, theta)
    d_phi = lambda u: diff(u, phi)
    scalar_lap = lambda u: spherical_laplacian(u, r, theta, phi)

    lap_r = \
        scalar_lap(u_r) \
        - 2 * u_r / r ** 2 \
        - 2 * d_theta(u_theta * sin(theta)) / (r ** 2 * sin(theta)) \
        - 2 * d_phi(u_phi) / (r ** 2 * sin(theta))

    lap_theta = \
        scalar_lap(u_theta) \
        - u_theta / (r ** 2 * sin(theta) ** 2) \
        + 2 * d_theta(u_r) / r ** 2 \
        - 2 * cos(theta) * d_phi(u_phi) / (r ** 2 * sin(theta) ** 2)

    lap_phi = \
        scalar_lap(u_phi) \
        - u_phi / (r ** 2 * sin(theta) ** 2) \
        + 2 * d_phi(u_r) / (r ** 2 * sin(theta)) \
        + 2 * cos(theta) * d_phi(u_theta) / (r ** 2 * sin(theta) ** 2)

    return lap_r, lap_theta, lap_phi



import numpy as np
import random
from numpy import isclose
import pytest
import matplotlib

matplotlib.use('Agg')  # use a non-GUI backend, so plots are not shown during testing

from neurodiffeq.neurodiffeq import safe_diff as diff
from neurodiffeq.networks import FCNN
from neurodiffeq.pde import DirichletControlPoint, NeumannControlPoint, Point, CustomBoundaryCondition
from neurodiffeq.pde import solve2D, solve2D_system, Monitor2D, make_animation
from neurodiffeq.pde import Solution
from neurodiffeq.pde import Solution2D
from neurodiffeq.generators import PredefinedGenerator, Generator2D
from neurodiffeq.conditions import DirichletBVP2D, DirichletBVP

import torch
import torch.nn as nn
import torch.optim as optim


@pytest.fixture(autouse=True)
def magic():
    torch.manual_seed(42)
    np.random.seed(42)


def test_monitor():
    laplace = lambda u, x, y: diff(u, x, order=2) + diff(u, y, order=2)
    bc = DirichletBVP2D(
        x_min=0, x_min_val=lambda y: torch.sin(np.pi * y),
        x_max=1, x_max_val=lambda y: 0,
        y_min=0, y_min_val=lambda x: 0,
        y_max=1, y_max_val=lambda x: 0
    )

    net = FCNN(n_input_units=2, hidden_units=(32, 32))
    with pytest.warns(FutureWarning):
        solve2D(
            pde=laplace, condition=bc, xy_min=(0, 0), xy_max=(1, 1),
            net=net, max_epochs=3,
            train_generator=Generator2D((32, 32), (0, 0), (1, 1), method='equally-spaced-noisy'),
            batch_size=64,
            monitor=Monitor2D(check_every=1, xy_min=(0, 0), xy_max=(1, 1))
        )


def test_train_generator():
    laplace = lambda u, x, y: diff(u, x, order=2) + diff(u, y, order=2)
    bc = DirichletBVP2D(
        x_min=0, x_min_val=lambda y: torch.sin(np.pi * y),
        x_max=1, x_max_val=lambda y: 0,
        y_min=0, y_min_val=lambda x: 0,
        y_max=1, y_max_val=lambda x: 0
    )

    net = FCNN(n_input_units=2, hidden_units=(32, 32))

    with pytest.raises(ValueError), pytest.warns(FutureWarning):
        solution_neural_net_laplace, _ = solve2D(
            pde=laplace, condition=bc,
            net=net, max_epochs=3, batch_size=64
        )


def test_laplace():
    laplace = lambda u, x, y: diff(u, x, order=2) + diff(u, y, order=2)
    bc = DirichletBVP2D(
        x_min=0, x_min_val=lambda y: torch.sin(np.pi * y),
        x_max=1, x_max_val=lambda y: 0,
        y_min=0, y_min_val=lambda x: 0,
        y_max=1, y_max_val=lambda x: 0,
    )

    net = FCNN(n_input_units=2, hidden_units=(32, 32))
    with pytest.warns(FutureWarning):
        solution_neural_net_laplace, loss_history = solve2D(
            pde=laplace, condition=bc, xy_min=(0, 0), xy_max=(1, 1),
            net=net, max_epochs=3,
            train_generator=Generator2D((32, 32), (0, 0), (1, 1), method='equally-spaced-noisy',
                                        xy_noise_std=(0.01, 0.01)),
            batch_size=64
        )
    assert isinstance(solution_neural_net_laplace, Solution2D)
    assert isinstance(loss_history, dict)
    keys = ['train_loss', 'valid_loss']
    for key in keys:
        assert key in loss_history
        assert isinstance(loss_history[key], list)
    assert len(loss_history[keys[0]]) == len(loss_history[keys[1]])


# def test_pde_system():
#     def _network_output_2input(net, xs, ys, ith_unit):
#         xys = torch.cat((xs, ys), 1)
#         nn_output = net(xys)
#         if ith_unit is not None:
#             return nn_output[:, ith_unit].reshape(-1, 1)
#         else:
#             return nn_output
#
#     class BCOnU(Condition):
#         """for u(x, y), impose u(x, -1) = u(x, 1) = 0; dudx(0, y) = dudy(L, y) = 0"""
#
#         def __init__(self, x_min, x_max, y_min, y_max):
#             super().__init__()
#             self.x_min = x_min
#             self.x_max = x_max
#             self.y_min = y_min
#             self.y_max = y_max
#
#         def enforce(self, net, x, y):
#             uxy = _network_output_2input(net, x, y, self.ith_unit)
#
#             x_ones = torch.ones_like(x, requires_grad=True)
#             x_ones_min = self.x_min * x_ones
#             x_ones_max = self.x_max * x_ones
#             uxminy = _network_output_2input(net, x_ones_min, y, self.ith_unit)
#             uxmaxy = _network_output_2input(net, x_ones_max, y, self.ith_unit)
#
#             x_tilde = (x - self.x_min) / (self.x_max - self.x_min)
#             y_tilde = (y - self.y_min) / (self.y_max - self.y_min)
#
#             return y_tilde * (1 - y_tilde) * (
#                     uxy - x_tilde * (self.x_max - self.x_min) * diff(uxminy, x_ones_min) \
#                     + 0.5 * x_tilde ** 2 * (self.x_max - self.x_min) * (
#                             diff(uxminy, x_ones_min) - diff(uxmaxy, x_ones_max)
#                     )
#             )
#
#     class BCOnP(Condition):
#         """for p(x, y), impose p(0, y) = p_max; p(L, y) = p_min"""
#
#         def __init__(self, x_min, x_max, p_x_min, p_x_max):
#             super().__init__()
#             self.x_min = x_min
#             self.x_max = x_max
#             self.p_x_min = p_x_min
#             self.p_x_max = p_x_max
#
#         def enforce(self, net, x, y):
#             uxy = _network_output_2input(net, x, y, self.ith_unit)
#             x_tilde = (x - self.x_min) / (self.x_max - self.x_min)
#
#             return (1 - x_tilde) * self.p_x_min + x_tilde * self.p_x_max \
#                    + x_tilde * (1 - x_tilde) * uxy
#
#     L = 2.0
#     mu = 1.0
#     P1, P2 = 1.0, 0.0
#     def poiseuille(u, v, p, x, y):
#         return [
#             mu * (diff(u, x, order=2) + diff(u, y, order=2)) - diff(p, x),
#             mu * (diff(v, x, order=2) + diff(v, y, order=2)) - diff(p, y),
#             diff(u, x) + diff(v, y)
#         ]
#     def zero_divergence(u, v, p, x, y):
#         return torch.sum( (diff(u, x) + diff(v, y))**2 )
#
#     bc_on_u = BCOnU(
#         x_min=0,
#         x_max=L,
#         y_min=-1,
#         y_max=1,
#     )
#     bc_on_v = DirichletBVP2D(
#         x_min=0, x_min_val=lambda y: 0,
#         x_max=L, x_max_val=lambda y: 0,
#         y_min=-1, y_min_val=lambda x: 0,
#         y_max=1, y_max_val=lambda x: 0
#     )
#     bc_on_p = BCOnP(
#         x_min=0,
#         x_max=L,
#         p_x_min=P1,
#         p_x_max=P2,
#     )
#     conditions = [bc_on_u, bc_on_v, bc_on_p]
#
#     nets = [
#         FCNN(n_input_units=2, hidden_units=(32, 32), actv=nn.Softplus)
#         for _ in range(3)
#     ]
#
#     # use one neural network for each dependent variable
#     solution_neural_net_poiseuille, _ = solve2D_system(
#         pde_system=poiseuille, conditions=conditions, xy_min=(0, -1), xy_max=(L, 1),
#         train_generator=Generator2D((32, 32), (0, -1), (L, 1), method='equally-spaced-noisy'),
#         max_epochs=300, batch_size=64, nets=nets, additional_loss_term=zero_divergence,
#         monitor=Monitor2D(check_every=10, xy_min=(0, -1), xy_max=(L, 1))
#     )
#
#     def solution_analytical_poiseuille(xs, ys):
#         us = (P1 - P2) / (L * 2 * mu) * (1 - ys ** 2)
#         vs = np.zeros_like(xs)
#         ps = P1 + (P2 - P1) * xs / L
#         return [us, vs, ps]
#
#     xs, ys = np.linspace(0, L, 101), np.linspace(-1, 1, 101)
#     xx, yy = np.meshgrid(xs, ys)
#     u_ana, v_ana, p_ana = solution_analytical_poiseuille(xx, yy)
#     u_net, v_net, p_net = solution_neural_net_poiseuille(xx, yy, to_numpy=True)
#
#     assert isclose(u_ana, u_net, atol=0.01).all()
#     assert isclose(v_ana, v_net, atol=0.01).all()
#     assert isclose(p_ana, p_net, atol=0.01).all()

def test_arbitrary_boundary():
    def solution_analytical_problem_c(x, y):
        return np.log(1 + x ** 2 + y ** 2)

    def gradient_solution_analytical_problem_c(x, y):
        return 2 * x / (1 + x ** 2 + y ** 2), 2 * y / (1 + x ** 2 + y ** 2),

    # creating control points for Dirichlet boundary conditions

    edge_length = 2.0 / np.sin(np.pi / 3) / 4
    points_on_each_edge = 11
    step_size = edge_length / (points_on_each_edge - 1)

    direction_theta = np.pi * 2 / 3
    left_turn_theta = np.pi * 1 / 3
    right_turn_theta = -np.pi * 2 / 3

    dirichlet_control_points_problem_c = []
    point_x, point_y = 0.0, -1.0
    for i_edge in range(6):
        for i_step in range(points_on_each_edge - 1):
            dirichlet_control_points_problem_c.append(
                DirichletControlPoint(
                    loc=(point_x, point_y),
                    val=solution_analytical_problem_c(point_x, point_y)
                )
            )
            point_x += step_size * np.cos(direction_theta)
            point_y += step_size * np.sin(direction_theta)
        direction_theta += left_turn_theta if (i_edge % 2 == 0) else right_turn_theta

    # dummy control points to form closed domain

    radius_circle = 1.0 / np.sin(np.pi / 6)
    center_circle_x = radius_circle * np.cos(np.pi / 6)
    center_circle_y = 0.0

    dirichlet_control_points_problem_c_dummy = []
    for theta in np.linspace(-np.pi * 5 / 6, np.pi * 5 / 6, 60):
        point_x = center_circle_x + radius_circle * np.cos(theta)
        point_y = center_circle_y + radius_circle * np.sin(theta)
        dirichlet_control_points_problem_c_dummy.append(
            DirichletControlPoint(
                loc=(point_x, point_y),
                val=solution_analytical_problem_c(point_x, point_y)
            )
        )

    # all Dirichlet control points

    dirichlet_control_points_problem_c_all = \
        dirichlet_control_points_problem_c + dirichlet_control_points_problem_c_dummy

    # creating control points for Neumann boundary condition

    edge_length = 2.0 / np.sin(np.pi / 3) / 4
    points_on_each_edge = 11
    step_size = edge_length / (points_on_each_edge - 1)

    normal_theta = np.pi / 6

    direction_theta = -np.pi * 1 / 3
    left_turn_theta = np.pi * 1 / 3
    right_turn_theta = -np.pi * 2 / 3

    neumann_control_points_problem_c = []
    point_x, point_y = 0.0, 1.0
    for i_edge in range(6):
        normal_x = np.cos(normal_theta)
        normal_y = np.sin(normal_theta)

        # skip the points on the "tip", their normal vector is undefined?
        point_x += step_size * np.cos(direction_theta)
        point_y += step_size * np.sin(direction_theta)

        for i_step in range(points_on_each_edge - 2):
            grad_x, grad_y = gradient_solution_analytical_problem_c(point_x, point_y)
            neumann_val = grad_x * normal_x + grad_y * normal_y
            neumann_control_points_problem_c.append(
                NeumannControlPoint(
                    loc=(point_x, point_y),
                    val=neumann_val,
                    normal_vector=(normal_x, normal_y)
                )
            )
            point_x += step_size * np.cos(direction_theta)
            point_y += step_size * np.sin(direction_theta)
        direction_theta += left_turn_theta if (i_edge % 2 == 0) else right_turn_theta
        normal_theta += left_turn_theta if (i_edge % 2 == 0) else right_turn_theta

    # dummy control points to form closed domain

    radius_circle = 1.0 / np.sin(np.pi / 6)
    center_circle_x = -radius_circle * np.cos(np.pi / 6)
    center_circle_y = 0.0

    neumann_control_points_problem_c_dummy = []
    for theta in np.linspace(np.pi * 1 / 6, np.pi * 11 / 6, 60):
        point_x = center_circle_x + radius_circle * np.cos(theta)
        point_y = center_circle_y + radius_circle * np.sin(theta)
        normal_x = np.cos(theta)
        normal_y = np.sin(theta)
        grad_x, grad_y = gradient_solution_analytical_problem_c(point_x, point_y)
        neumann_val = grad_x * normal_x + grad_y * normal_y
        neumann_control_points_problem_c_dummy.append(
            NeumannControlPoint(
                loc=(point_x, point_y),
                val=neumann_val,
                normal_vector=(normal_x, normal_y)
            )
        )

    # all Neumann control points

    neumann_control_points_problem_c_all = \
        neumann_control_points_problem_c + neumann_control_points_problem_c_dummy

    cbc_problem_c = CustomBoundaryCondition(
        center_point=Point(loc=(0.0, 0.0)),
        dirichlet_control_points=dirichlet_control_points_problem_c_all,
        neumann_control_points=neumann_control_points_problem_c_all
    )

    def get_grid(x_from_to, y_from_to, x_n_points=100, y_n_points=100, as_tensor=False):
        x_from, x_to = x_from_to
        y_from, y_to = y_from_to
        if as_tensor:
            x = torch.linspace(x_from, x_to, x_n_points)
            y = torch.linspace(y_from, y_to, y_n_points)
            return torch.meshgrid(x, y)
        else:
            x = np.linspace(x_from, x_to, x_n_points)
            y = np.linspace(y_from, y_to, y_n_points)
            return np.meshgrid(x, y)

    def to_np(tensor):
        return tensor.detach().cpu().numpy()

    xx_train, yy_train = get_grid(
        x_from_to=(-1, 1), y_from_to=(-1, 1),
        x_n_points=28, y_n_points=28,
        as_tensor=True
    )
    is_in_domain_train = cbc_problem_c.in_domain(xx_train, yy_train)
    xx_train, yy_train = to_np(xx_train), to_np(yy_train)
    xx_train, yy_train = xx_train[is_in_domain_train], yy_train[is_in_domain_train]
    train_gen = PredefinedGenerator(xx_train, yy_train)

    xx_valid, yy_valid = get_grid(
        x_from_to=(-1, 1), y_from_to=(-1, 1),
        x_n_points=10, y_n_points=10,
        as_tensor=True
    )
    is_in_domain_valid = cbc_problem_c.in_domain(xx_valid, yy_valid)
    xx_valid, yy_valid = to_np(xx_valid), to_np(yy_valid)
    xx_valid, yy_valid = xx_valid[is_in_domain_valid], yy_valid[is_in_domain_valid]
    valid_gen = PredefinedGenerator(xx_valid, yy_valid)

    def rmse(u, x, y):
        true_u = torch.log(1 + x ** 2 + y ** 2)
        return torch.mean((u - true_u) ** 2) ** 0.5

    # nabla^2 psi(x, y) = (e^(-x))(x-2+y^3+6y)
    def de_problem_c(u, x, y):
        return diff(u, x, order=2) + diff(u, y, order=2) + torch.exp(u) - 1.0 - x ** 2 - y ** 2 - 4.0 / (
                1.0 + x ** 2 + y ** 2) ** 2

    # fully connected network with one hidden layer (100 hidden units with ELU activation)
    net = FCNN(n_input_units=2, hidden_units=(100, 100), actv=nn.ELU)
    adam = optim.Adam(params=net.parameters(), lr=0.001)

    # train on 28 X 28 grid
    with pytest.warns(FutureWarning):
        solution_neural_net_problem_c, history_problem_c = solve2D(
            pde=de_problem_c, condition=cbc_problem_c,
            xy_min=(-1, -1), xy_max=(1, 1),
            train_generator=train_gen, valid_generator=valid_gen,
            net=net, max_epochs=1, batch_size=128, optimizer=adam,
            monitor=Monitor2D(check_every=1, xy_min=(-1, -1), xy_max=(1, 1), valid_generator=valid_gen),
            metrics={'rmse': rmse}
        )

    xs = torch.tensor([p.loc[0] for p in dirichlet_control_points_problem_c], requires_grad=True).reshape(-1, 1)
    ys = torch.tensor([p.loc[1] for p in dirichlet_control_points_problem_c], requires_grad=True).reshape(-1, 1)
    us = solution_neural_net_problem_c(xs, ys, to_numpy=True)
    true_us = solution_analytical_problem_c(to_np(xs), to_np(ys))
    assert isclose(us, true_us, atol=1e-4).all()

    xs = torch.tensor([p.loc[0] for p in neumann_control_points_problem_c], requires_grad=True).reshape(-1, 1)
    ys = torch.tensor([p.loc[1] for p in neumann_control_points_problem_c], requires_grad=True).reshape(-1, 1)
    us = solution_neural_net_problem_c(xs, ys)
    nxs = torch.tensor([p.normal_vector[0] for p in neumann_control_points_problem_c]).reshape(-1, 1)
    nys = torch.tensor([p.normal_vector[1] for p in neumann_control_points_problem_c]).reshape(-1, 1)
    normal_derivative = to_np(nxs * diff(us, xs) + nys * diff(us, ys)).flatten()
    true_normal_derivative = np.array([p.val for p in neumann_control_points_problem_c])
    assert isclose(normal_derivative, true_normal_derivative, atol=1e-2).all()


def test_solution():
    x_grids = 7
    y_grids = 11
    xy_grids = (x_grids, y_grids)
    N_SAMPLES = x_grids * y_grids
    x0, y0 = random.random(), random.random()
    x1, y1 = random.random() + 1, random.random() + 1
    generator = Generator2D(xy_grids, xy_min=(x0, y0), xy_max=(x1, y1), method='equally-spaced')

    u00, u01, u10, u11 = random.random(), random.random(), random.random(), random.random()
    v00, v01, v10, v11 = random.random(), random.random(), random.random(), random.random()

    def get_single_boundary_func(z0, w0, z1, w1):
        net = FCNN(1, 1)
        condition = DirichletBVP(z0, w0, z1, w1)

        def boundary_func(z):
            return condition.enforce(net, z)

        return boundary_func

    def get_all_boundary_funcs(w00, w01, w10, w11):
        wx0 = get_single_boundary_func(y0, w00, y1, w01)
        wx1 = get_single_boundary_func(y0, w10, y1, w11)
        wy0 = get_single_boundary_func(x0, w00, x1, w10)
        wy1 = get_single_boundary_func(x0, w01, x1, w11)
        return wx0, wx1, wy0, wy1

    ux0, ux1, uy0, uy1 = get_all_boundary_funcs(u00, u01, u10, u11)
    vx0, vx1, vy0, vy1 = get_all_boundary_funcs(v00, v01, v10, v11)

    def get_solution(use_single: bool) -> Solution2D:
        conditions = [
            DirichletBVP2D(x0, ux0, x1, ux1, y0, uy0, y1, uy1),
            DirichletBVP2D(x0, vx0, x1, vx1, y0, vy0, y1, vy1),
        ]
        if use_single:
            net = FCNN(2, 2)
            for i, cond in enumerate(conditions):
                with pytest.warns(DeprecationWarning):
                    cond.set_impose_on(i)
            return Solution2D(net, conditions)
        else:
            nets = [FCNN(2, 1), FCNN(2, 1)]
            return Solution2D(nets, conditions)

    def check_output(uv, shape, type, msg=""):
        msg += " "
        assert isinstance(uv, (list, tuple)), msg + "returned type is not a list"
        assert len(uv) == 2, msg + "returned length is not 2"
        assert isinstance(uv[0], type) and isinstance(uv[1], type), msg + f"returned element is not {type}"
        u, v = uv
        assert u.shape == shape and v.shape == shape, msg + f"returned element shape is not {shape}"
        u, v = u.reshape(*xy_grids), v.reshape(*xy_grids)
        x = torch.linspace(x0, x1, steps=x_grids, requires_grad=True).reshape(-1, 1)
        y = torch.linspace(y0, y1, steps=y_grids, requires_grad=True).reshape(-1, 1)

        if type == torch.Tensor:
            check_close = lambda a, b: torch.isclose(a, b).all()
        elif type == np.ndarray:
            check_close = lambda a, b: np.isclose(a, b.cpu().detach().numpy()).all
        else:
            raise ValueError(f"Unrecognized type={type}")

        assert check_close(u[0, :], torch.flatten(ux0(y))), msg + "u on x0 not satisfied"
        assert check_close(u[-1, :], torch.flatten(ux1(y))), msg + "u on x1 not satisfied"
        assert check_close(u[:, 0], torch.flatten(uy0(x))), msg + "u on y0 not satisfied"
        assert check_close(u[:, -1], torch.flatten(uy1(x))), msg + "u on y1 not satisfied"

        assert check_close(v[0, :], torch.flatten(vx0(y))), msg + "v on x0 not satisfied"
        assert check_close(v[-1, :], torch.flatten(vx1(y))), msg + "v on x1 not satisfied"
        assert check_close(v[:, 0], torch.flatten(vy0(x))), msg + "v on y0 not satisfied"
        assert check_close(v[:, -1], torch.flatten(vy1(x))), msg + "v on y1 not satisfied"

    for use_single in [True, False]:
        solution = get_solution(use_single=use_single)
        xs, ys = generator.get_examples()
        us = solution(xs, ys)
        check_output(us, shape=(N_SAMPLES,), type=torch.Tensor, msg=f"[use_single={use_single}]")
        with pytest.warns(FutureWarning):
            us = solution(xs, ys, as_type='np')
        check_output(us, shape=(N_SAMPLES,), type=np.ndarray, msg=f"[use_single={use_single}]")
        us = solution(xs, ys, to_numpy=True)
        check_output(us, shape=(N_SAMPLES,), type=np.ndarray, msg=f"[use_single={use_single}]")

        xs, ys = xs.reshape(-1, 1), ys.reshape(-1, 1)
        us = solution(xs, ys)
        check_output(us, shape=(N_SAMPLES, 1), type=torch.Tensor, msg=f"[use_single={use_single}]")
        with pytest.warns(FutureWarning):
            us = solution(xs, ys, as_type='np')
        check_output(us, shape=(N_SAMPLES, 1), type=np.ndarray, msg=f"[use_single={use_single}]")
        us = solution(xs, ys, to_numpy=True)
        check_output(us, shape=(N_SAMPLES, 1), type=np.ndarray, msg=f"[use_single={use_single}]")



from math import pi as PI
import torch
from torch import nn, optim
from neurodiffeq.neurodiffeq import unsafe_diff as diff
from neurodiffeq.networks import FCNN
from neurodiffeq.temporal import generator_1dspatial, generator_temporal
from neurodiffeq.temporal import generator_2dspatial_segment, generator_2dspatial_rectangle
from neurodiffeq.temporal import FirstOrderInitialCondition, BoundaryCondition
from neurodiffeq.temporal import SingleNetworkApproximator1DSpatialTemporal, SingleNetworkApproximator2DSpatial, SingleNetworkApproximator2DSpatialTemporal
from neurodiffeq.temporal import Monitor1DSpatialTemporal, Monitor2DSpatial, Monitor2DSpatialTemporal
from neurodiffeq.temporal import _train_1dspatial_temporal, _valid_1dspatial_temporal, _solve_1dspatial_temporal
from neurodiffeq.temporal import _train_2dspatial_temporal, _valid_2dspatial_temporal, _solve_2dspatial_temporal
from neurodiffeq.temporal import _train_2dspatial, _valid_2dspatial, _solve_2dspatial
import matplotlib
matplotlib.use('Agg') # use a non-GUI backend, so plots are not shown during testing


def test_generator_1dspatial():
    s_gen = generator_1dspatial(size=32, x_min=-4, x_max=2, random=False)
    for _ in range(3):
        x = next(s_gen)
        assert x.shape == torch.Size([32])
        assert (x >= -4).all()
        assert (x <= 2).all()
        assert not x.requires_grad
    assert (x == next(s_gen)).all()

    s_gen = generator_1dspatial(size=32, x_min=-4, x_max=2, random=True)
    for _ in range(3):
        x = next(s_gen)
        assert x.shape == torch.Size([32])
        assert (x >= -4).all()
        assert (x <= 2).all()
        assert not x.requires_grad
    assert not (x == next(s_gen)).all()


def test_generator_2dspatial_segment():
    s_gen = generator_2dspatial_segment(size=32, start=(4., 2.), end=(-2., -4.), random=False)
    for _ in range(3):
        x, y = next(s_gen)
        assert x.shape == torch.Size([32]) and y.shape == torch.Size([32])
        assert (x - y - 2.).isclose(torch.zeros_like(x)).all()
        assert not x.requires_grad
    x_, y_ = next(s_gen)
    assert (x == x_).all() and (y == y_).all()

    s_gen = generator_2dspatial_segment(size=32, start=(4., 2.), end=(-2., -4.), random=True)
    for _ in range(3):
        x, y = next(s_gen)
        assert x.shape == torch.Size([32]) and y.shape == torch.Size([32])
        assert (x - y - 2.).isclose(torch.zeros_like(x), atol=1e-6).all()
        assert not x.requires_grad
    x_, y_ = next(s_gen)
    assert not (x == x_).all()
    assert not (y == y_).all()


def test_generator_2dspatial_rectangle():
    s_gen = generator_2dspatial_rectangle(size=(8, 8), x_min=-2., x_max=4., y_min=-4., y_max=2., random=False)
    for _ in range(3):
        x, y = next(s_gen)
        assert x.shape == torch.Size([64]) and y.shape == torch.Size([64])
        assert (x >= -2.).all() and (x <= 4.).all()
        assert (y >= -4.).all() and (y <= 2.).all()
        assert not x.requires_grad
    x_, y_ = next(s_gen)
    assert (x == x_).all() and (y == y_).all()

    s_gen = generator_2dspatial_rectangle(size=(8, 8), x_min=-2., x_max=4., y_min=-4., y_max=2., random=True)
    for _ in range(3):
        x, y = next(s_gen)
        assert x.shape == torch.Size([64]) and y.shape == torch.Size([64])
        assert (x >= -2.).all() and (x <= 4.).all()
        assert (y >= -4.).all() and (y <= 2.).all()
        assert not x.requires_grad
    x_, y_ = next(s_gen)
    assert not (x == x_).all()
    assert not (y == y_).all()


def test_generator_temporal():
    t_gen = generator_temporal(size=32, t_min=0, t_max=42, random=False)
    for _ in range(3):
        t = next(t_gen)
        assert t.shape == torch.Size([32])
        assert (t >= 0).all()
        assert (t <= 42).all()
        assert not t.requires_grad
    assert (t == next(t_gen)).all()

    t_gen = generator_temporal(size=32, t_min=0, t_max=42, random=True)
    for _ in range(3):
        t = next(t_gen)
        assert t.shape == torch.Size([32])
        assert (t >= 0).all()
        assert (t <= 42).all()
        assert not t.requires_grad
    assert not (t == next(t_gen)).all()


def test_first_order_initial_condition():
    initial_condition = FirstOrderInitialCondition(u0=lambda x: torch.sin(x))
    x = torch.linspace(0, 1, 32)
    assert (initial_condition.u0(x) == torch.sin(x)).all()


def test_boundary_condition():
    def points_gen():
        while True:
            yield torch.tensor([0.])
    boundary_condition = BoundaryCondition(
        form=lambda u, x, t: t,
        points_generator=points_gen()
    )
    x = next(boundary_condition.points_generator)
    assert (x == torch.tensor([0.])).all()

    t_gen = generator_temporal(size=32, t_min=0, t_max=42, random=True)
    t = next(t_gen)
    xt = torch.cartesian_prod(x, t)
    xx, tt = xt[:, 0], xt[:, 1]
    def dummy_u(x, t):
        return t
    uu = dummy_u(xx, tt)
    assert (boundary_condition.form(uu, xx, tt) == tt).all()


def test_single_network_approximator_1dspatial_temporal():
    DIFFUSIVITY, X_MIN, X_MAX, T_MIN, T_MAX = 0.3, 0.0, 2.0, 0.0, 3.0

    def heat_equation_1d(u, x, t):
        return diff(u, t) - DIFFUSIVITY * diff(u, x, order=2)

    initial_condition = FirstOrderInitialCondition(u0=lambda x: torch.sin(PI * x))

    def points_gen_lo():
        while True:
            yield torch.tensor([X_MIN])
    dirichlet_boundary_lo = BoundaryCondition(
        form=lambda u, x, t: torch.zeros_like(u),
        points_generator=points_gen_lo()
    )
    def points_gen_hi():
        while True:
            yield torch.tensor([X_MAX])
    dirichlet_boundary_hi = BoundaryCondition(
        form=lambda u, x, t: torch.zeros_like(u),
        points_generator=points_gen_hi()
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator1DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_1d,
        initial_condition=initial_condition,
        boundary_conditions=[dirichlet_boundary_lo, dirichlet_boundary_hi]
    )

    xx, tt = torch.rand(16), torch.rand(16)
    assert fcnn_approximator(xx, tt).shape == torch.Size([16])
    assert next(fcnn_approximator.parameters()).shape == torch.Size([32, 2])
    x, t = torch.rand(4), torch.rand(4)
    xt = torch.cartesian_prod(x, t)
    xx = torch.squeeze(xt[:, 0])
    xx.requires_grad = True
    tt = torch.squeeze(xt[:, 1])
    tt.requires_grad = True
    assert fcnn_approximator.calculate_loss(xx, tt, x, t).shape == torch.Size([])
    def dummy_mse(uu, xx, tt):
        return torch.mean((uu - (xx+tt))**2)
    metrics = {'dummy_mse': dummy_mse}
    assert fcnn_approximator.calculate_metrics(xx, tt, x, t, metrics)['dummy_mse'].shape == torch.Size([])
    xx, tt = torch.rand(16), torch.zeros(16)
    assert fcnn_approximator(xx, tt).isclose(torch.sin(PI * xx)).all()


def test_single_network_approximator_2dspatial():
    def laplace_2d(u, xx, yy):
        return diff(u, xx, order=2) + diff(u, yy, order=2)

    def analytical_solution(xx, yy):
        return torch.sin(PI * yy) * torch.sinh(PI * (1 - xx)) / torch.sinh(torch.ones_like(xx) * PI)

    metrics = {}

    def rmse(uu, xx, yy):
        error = uu - analytical_solution(xx, yy)
        return torch.mean(error ** 2) ** 0.5

    metrics['rmse'] = rmse

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y: u - torch.sin(PI * y),
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(0.0, 1.0))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(1.0, 0.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 1.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(1.0, 1.0))
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatial(
        single_network=fcnn,
        pde=laplace_2d,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )

    xx, yy = torch.rand(16), torch.rand(16)
    assert fcnn_approximator(xx, yy).shape == torch.Size([16])
    assert next(fcnn_approximator.parameters()).shape == torch.Size([32, 2])
    xx.requires_grad = True
    yy.requires_grad = True
    assert fcnn_approximator.calculate_loss(xx, yy).shape == torch.Size([])
    assert fcnn_approximator.calculate_metrics(xx, yy, metrics)['rmse'].shape == torch.Size([])


def test_single_network_approximator_2dspatial_temporal():
    DIFFUSIVITY = 0.3
    X_MIN, X_MAX = -1.0, 1.0
    Y_MIN, Y_MAX = -1.0, 1.0

    def heat_equation_2d(u, x, y, t):
        left = diff(u, t) - DIFFUSIVITY * (diff(u, x, order=2) + diff(u, y, order=2))
        right = -torch.exp(-t) * ((X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN) - 2 * DIFFUSIVITY * (
                    (Y_MAX - y) * (y - Y_MIN) + (X_MAX - x) * (x - X_MIN)))
        return left - right

    def analytical_solution(xx, yy, tt):
        return torch.exp(-tt) * (X_MAX - xx) * (xx - X_MIN) * (Y_MAX - yy) * (yy - Y_MIN)

    def rmse(uu, xx, yy, tt):
        error = uu - analytical_solution(xx, yy, tt)
        return torch.mean(error ** 2) ** 0.5

    metrics = {'rmse': rmse}

    def u0(x, y):
        return (X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN)

    initial_condition = FirstOrderInitialCondition(u0=u0)

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MIN, Y_MAX))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MAX, Y_MIN), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MAX), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MAX, Y_MIN))
    )

    fcnn = FCNN(
        n_input_units=3,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_2d,
        initial_condition=initial_condition,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )

    xx, yy, tt = torch.rand(16), torch.rand(16), torch.rand(16)
    assert fcnn_approximator(xx, yy, tt).shape == torch.Size([16])
    assert next(fcnn_approximator.parameters()).shape == torch.Size([32, 3])
    x, y, t = torch.rand(4), torch.rand(4), torch.rand(4)
    xt = torch.cartesian_prod(x, t)
    yt = torch.cartesian_prod(y, t)
    xx = torch.squeeze(xt[:, 0])
    xx.requires_grad = True
    yy = torch.squeeze(yt[:, 0])
    yy.requires_grad = True
    tt = torch.squeeze(yt[:, 1])
    tt.requires_grad = True
    assert fcnn_approximator.calculate_loss(xx, yy, tt, x, y, t).shape == torch.Size([])
    assert fcnn_approximator.calculate_metrics(xx, yy, tt, x, y, t, metrics)['rmse'].shape == torch.Size([])
    xx, yy, tt = torch.rand(16), torch.rand(16), torch.zeros(16)
    assert fcnn_approximator(xx, yy, tt).isclose(u0(xx, yy)).all()


def test_monitor_1dspatial_temporal():
    DIFFUSIVITY, X_MIN, X_MAX, T_MIN, T_MAX = 0.3, 0.0, 2.0, 0.0, 3.0

    def heat_equation_1d(u, x, t):
        return diff(u, t) - DIFFUSIVITY * diff(u, x, order=2)

    initial_condition = FirstOrderInitialCondition(u0=lambda x: torch.sin(PI * x))

    def points_gen_lo():
        while True:
            yield torch.tensor([X_MIN])

    dirichlet_boundary_lo = BoundaryCondition(
        form=lambda u, x, t: torch.zeros_like(u),
        points_generator=points_gen_lo()
    )

    def points_gen_hi():
        while True:
            yield torch.tensor([X_MAX])

    dirichlet_boundary_hi = BoundaryCondition(
        form=lambda u, x, t: torch.zeros_like(u),
        points_generator=points_gen_hi()
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator1DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_1d,
        initial_condition=initial_condition,
        boundary_conditions=[dirichlet_boundary_lo, dirichlet_boundary_hi]
    )

    dummy_history = {
        'train_loss': [100, 10, 1],
        'valid_loss': [200, 20, 2],
        'train_rmse': [1, 0.1, 0.01],
        'valid_rmse': [2, 0.2, 0.02]
    }

    monitor = Monitor1DSpatialTemporal(
        check_on_x=torch.linspace(X_MIN, X_MAX, 32),
        check_on_t=torch.linspace(T_MIN, T_MAX, 4),
        check_every=10
    )
    monitor.check(fcnn_approximator, dummy_history)
    monitor.check(fcnn_approximator, dummy_history)


def test_monitor_2dspatial():
    def laplace_2d(u, xx, yy):
        return diff(u, xx, order=2) + diff(u, yy, order=2)

    def analytical_solution(xx, yy):
        return torch.sin(PI * yy) * torch.sinh(PI * (1 - xx)) / torch.sinh(torch.ones_like(xx) * PI)

    metrics = {}

    def rmse(uu, xx, yy):
        error = uu - analytical_solution(xx, yy)
        return torch.mean(error ** 2) ** 0.5

    metrics['rmse'] = rmse

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y: u - torch.sin(PI * y),
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(0.0, 1.0))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(1.0, 0.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 1.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(1.0, 1.0))
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatial(
        single_network=fcnn,
        pde=laplace_2d,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )

    monitor = Monitor2DSpatial(
        check_on_x=torch.linspace(0.0, 1.0, 32),
        check_on_y=torch.linspace(0.0, 1.0, 32),
        check_every=10
    )

    dummy_history = {
        'train_loss': [100, 10, 1],
        'valid_loss': [200, 20, 2],
        'train_rmse': [1, 0.1, 0.01],
        'valid_rmse': [2, 0.2, 0.02]
    }

    monitor.check(fcnn_approximator, dummy_history)
    monitor.check(fcnn_approximator, dummy_history)


def test_monitor_2dspatial_temporal():
    DIFFUSIVITY = 0.3
    X_MIN, X_MAX = -1.0, 1.0
    Y_MIN, Y_MAX = -1.0, 1.0
    T_MIN, T_MAX = 0.0, 6.0

    def heat_equation_2d(u, x, y, t):
        left = diff(u, t) - DIFFUSIVITY * (diff(u, x, order=2) + diff(u, y, order=2))
        right = -torch.exp(-t) * ((X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN) - 2 * DIFFUSIVITY * (
                    (Y_MAX - y) * (y - Y_MIN) + (X_MAX - x) * (x - X_MIN)))
        return left - right

    def analytical_solution(xx, yy, tt):
        return torch.exp(-tt) * (X_MAX - xx) * (xx - X_MIN) * (Y_MAX - yy) * (yy - Y_MIN)

    def rmse(uu, xx, yy, tt):
        error = uu - analytical_solution(xx, yy, tt)
        return torch.mean(error ** 2) ** 0.5

    metrics = {'rmse': rmse}

    def u0(x, y):
        return (X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN)

    initial_condition = FirstOrderInitialCondition(u0=u0)

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MIN, Y_MAX))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MAX, Y_MIN), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MAX), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MAX, Y_MIN))
    )

    fcnn = FCNN(
        n_input_units=3,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_2d,
        initial_condition=initial_condition,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )

    monitor = Monitor2DSpatialTemporal(
        check_on_x=torch.linspace(X_MIN, X_MAX, 32),
        check_on_y=torch.linspace(Y_MIN, Y_MAX, 32),
        check_on_t=torch.linspace(T_MIN, T_MAX, 4),
        check_every=10
    )

    dummy_history = {
        'train_loss': [100, 10, 1],
        'valid_loss': [200, 20, 2],
        'train_rmse': [1, 0.1, 0.01],
        'valid_rmse': [2, 0.2, 0.02]
    }

    monitor.check(fcnn_approximator, dummy_history)
    monitor.check(fcnn_approximator, dummy_history)


def test__train_1dspatial_temporal():
    DIFFUSIVITY, X_MIN, X_MAX, T_MIN, T_MAX = 0.3, 0.0, 2.0, 0.0, 3.0

    def heat_equation_1d(u, x, t):
        return diff(u, t) - DIFFUSIVITY * diff(u, x, order=2)

    initial_condition = FirstOrderInitialCondition(u0=lambda x: torch.sin(PI * x))

    def points_gen_lo():
        while True:
            yield torch.tensor([X_MIN])

    dirichlet_boundary_lo = BoundaryCondition(
        form=lambda u, x, t: torch.zeros_like(u),
        points_generator=points_gen_lo()
    )

    def points_gen_hi():
        while True:
            yield torch.tensor([X_MAX])

    dirichlet_boundary_hi = BoundaryCondition(
        form=lambda u, x, t: torch.zeros_like(u),
        points_generator=points_gen_hi()
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator1DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_1d,
        initial_condition=initial_condition,
        boundary_conditions=[dirichlet_boundary_lo, dirichlet_boundary_hi]
    )

    s_gen = generator_1dspatial(size=32, x_min=X_MIN, x_max=X_MAX)
    t_gen = generator_temporal(size=32, t_min=T_MIN, t_max=T_MAX)

    adam = optim.Adam(fcnn_approximator.parameters())

    def dummy_mse(uu, xx, tt):
        return torch.mean((uu - (xx+tt))**2)
    metrics = {'dummy_mse': dummy_mse}

    train_epoch_loss, train_epoch_metrics = _train_1dspatial_temporal(s_gen, t_gen, fcnn_approximator, adam, metrics, shuffle=True, batch_size=100)
    assert train_epoch_loss > 0
    assert train_epoch_metrics['dummy_mse'] > 0


def test__valid_1dspatial_temporal():
    DIFFUSIVITY, X_MIN, X_MAX, T_MIN, T_MAX = 0.3, 0.0, 2.0, 0.0, 3.0

    def heat_equation_1d(u, x, t):
        return diff(u, t) - DIFFUSIVITY * diff(u, x, order=2)

    initial_condition = FirstOrderInitialCondition(u0=lambda x: torch.sin(PI * x))

    def points_gen_lo():
        while True:
            yield torch.tensor([X_MIN])

    dirichlet_boundary_lo = BoundaryCondition(
        form=lambda u, x, t: torch.zeros_like(u),
        points_generator=points_gen_lo()
    )

    def points_gen_hi():
        while True:
            yield torch.tensor([X_MAX])

    dirichlet_boundary_hi = BoundaryCondition(
        form=lambda u, x, t: torch.zeros_like(u),
        points_generator=points_gen_hi()
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator1DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_1d,
        initial_condition=initial_condition,
        boundary_conditions=[dirichlet_boundary_lo, dirichlet_boundary_hi]
    )

    s_gen = generator_1dspatial(size=32, x_min=X_MIN, x_max=X_MAX)
    t_gen = generator_temporal(size=32, t_min=T_MIN, t_max=T_MAX)

    def dummy_mse(uu, xx, tt):
        return torch.mean((uu - (xx + tt)) ** 2)

    metrics = {'dummy_mse': dummy_mse}

    valid_epoch_loss, valid_epoch_metrics = _valid_1dspatial_temporal(s_gen, t_gen, fcnn_approximator, metrics)
    assert valid_epoch_loss > 0
    assert valid_epoch_metrics['dummy_mse'] > 0


def test__solve_1dspatial_temporal():
    DIFFUSIVITY, X_MIN, X_MAX, T_MIN, T_MAX = 0.3, 0.0, 2.0, 0.0, 6.0

    def heat_equation_1d(u, x, t):
        return diff(u, t) - DIFFUSIVITY * diff(u, x, order=2)

    def analytical_solution(xx, tt):
        return torch.sin(PI * xx / X_MAX) * torch.exp(-DIFFUSIVITY * PI ** 2 * tt / X_MAX ** 2)

    def rmse(uu, xx, tt):
        error = uu - analytical_solution(xx, tt)
        return torch.mean(error ** 2) ** 0.5

    metrics = {'rmse': rmse}

    initial_condition = FirstOrderInitialCondition(u0=lambda x: torch.sin(PI * x / X_MAX))

    # a generator seems too much trouble for 1D boundary, but may be more flexible for 2D boundary
    def points_gen_lo():
        while True:
            yield torch.tensor([X_MIN])

    dirichlet_boundary_lo = BoundaryCondition(
        form=lambda u, x, t: u,
        points_generator=points_gen_lo()
    )

    def points_gen_hi():
        while True:
            yield torch.tensor([X_MAX])

    dirichlet_boundary_hi = BoundaryCondition(
        form=lambda u, x, t: u,
        points_generator=points_gen_hi()
    )

    train_gen_spatial = generator_1dspatial(size=32, x_min=X_MIN, x_max=X_MAX)
    train_gen_temporal = generator_temporal(size=32, t_min=T_MIN, t_max=T_MAX)
    valid_gen_spatial = generator_1dspatial(size=32, x_min=X_MIN, x_max=X_MAX, random=False)
    valid_gen_temporal = generator_temporal(size=32, t_min=T_MIN, t_max=T_MAX, random=False)
    monitor = Monitor1DSpatialTemporal(
        check_on_x=torch.linspace(X_MIN, X_MAX, 32),
        check_on_t=torch.linspace(T_MIN, T_MAX, 4),
        check_every=10
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator1DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_1d,
        initial_condition=initial_condition,
        boundary_conditions=[dirichlet_boundary_lo, dirichlet_boundary_hi]
    )
    adam = optim.Adam(fcnn_approximator.parameters())

    heat_equation_1d_solution, _ = _solve_1dspatial_temporal(
        train_generator_spatial=train_gen_spatial,
        train_generator_temporal=train_gen_temporal,
        valid_generator_spatial=valid_gen_spatial,
        valid_generator_temporal=valid_gen_temporal,
        approximator=fcnn_approximator,
        optimizer=adam,
        batch_size=64,
        max_epochs=1,
        shuffle=True,
        metrics=metrics,
        monitor=monitor
    )

    xx, tt = torch.rand(16), torch.rand(16)
    assert heat_equation_1d_solution(xx, tt).shape == torch.Size([16])
    xx, tt = torch.rand(16), torch.zeros(16)
    assert fcnn_approximator(xx, tt).isclose(torch.sin(PI * xx / X_MAX)).all()


def test__train_2dspatial():
    def laplace_2d(u, xx, yy):
        return diff(u, xx, order=2) + diff(u, yy, order=2)

    def analytical_solution(xx, yy):
        return torch.sin(PI * yy) * torch.sinh(PI * (1 - xx)) / torch.sinh(torch.ones_like(xx) * PI)

    metrics = {}

    def rmse(uu, xx, yy):
        error = uu - analytical_solution(xx, yy)
        return torch.mean(error ** 2) ** 0.5

    metrics['rmse'] = rmse

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y: u - torch.sin(PI * y),
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(0.0, 1.0))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(1.0, 0.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 1.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(1.0, 1.0))
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatial(
        single_network=fcnn,
        pde=laplace_2d,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )

    train_gen_spatial = generator_2dspatial_rectangle(
        size=(8, 8), x_min=0.0, x_max=1.0, y_min=0.0, y_max=1.0
    )

    adam = optim.Adam(fcnn_approximator.parameters())

    train_epoch_loss, train_epoch_metrics = _train_2dspatial(train_gen_spatial, None, fcnn_approximator, adam, metrics,
                                                             shuffle=True, batch_size=100)
    assert train_epoch_loss > 0
    assert train_epoch_metrics['rmse'] > 0


def test__valid_2dspatial():
    def laplace_2d(u, xx, yy):
        return diff(u, xx, order=2) + diff(u, yy, order=2)

    def analytical_solution(xx, yy):
        return torch.sin(PI * yy) * torch.sinh(PI * (1 - xx)) / torch.sinh(torch.ones_like(xx) * PI)

    metrics = {}

    def rmse(uu, xx, yy):
        error = uu - analytical_solution(xx, yy)
        return torch.mean(error ** 2) ** 0.5

    metrics['rmse'] = rmse

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y: u - torch.sin(PI * y),
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(0.0, 1.0))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(1.0, 0.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 1.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(1.0, 1.0))
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatial(
        single_network=fcnn,
        pde=laplace_2d,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )

    valid_gen_spatial = generator_2dspatial_rectangle(
        size=(8, 8), x_min=0.0, x_max=1.0, y_min=0.0, y_max=1.0
    )

    adam = optim.Adam(fcnn_approximator.parameters())

    valid_epoch_loss, valid_epoch_metrics = _valid_2dspatial(valid_gen_spatial, None, fcnn_approximator, metrics)
    assert valid_epoch_loss > 0
    assert valid_epoch_metrics['rmse'] > 0


def test__train_2dspatial_temporal():
    DIFFUSIVITY = 0.3
    X_MIN, X_MAX = -1.0, 1.0
    Y_MIN, Y_MAX = -1.0, 1.0
    T_MIN, T_MAX = 0.0, 6.0

    def heat_equation_2d(u, x, y, t):
        left = diff(u, t) - DIFFUSIVITY * (diff(u, x, order=2) + diff(u, y, order=2))
        right = -torch.exp(-t) * ((X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN) - 2 * DIFFUSIVITY * (
                (Y_MAX - y) * (y - Y_MIN) + (X_MAX - x) * (x - X_MIN)))
        return left - right

    def analytical_solution(xx, yy, tt):
        return torch.exp(-tt) * (X_MAX - xx) * (xx - X_MIN) * (Y_MAX - yy) * (yy - Y_MIN)

    def rmse(uu, xx, yy, tt):
        error = uu - analytical_solution(xx, yy, tt)
        return torch.mean(error ** 2) ** 0.5

    metrics = {'rmse': rmse}

    def u0(x, y):
        return (X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN)

    initial_condition = FirstOrderInitialCondition(u0=u0)

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MIN, Y_MAX))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MAX, Y_MIN), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MAX), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MAX, Y_MIN))
    )

    fcnn = FCNN(
        n_input_units=3,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_2d,
        initial_condition=initial_condition,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )

    train_gen_spatial = generator_2dspatial_rectangle(
        size=(8, 8), x_min=X_MIN, x_max=X_MAX, y_min=Y_MIN, y_max=Y_MAX
    )
    train_gen_temporal = generator_temporal(size=8, t_min=T_MIN, t_max=T_MAX)

    adam = optim.Adam(fcnn_approximator.parameters())

    train_epoch_loss, train_epoch_metrics = _train_2dspatial_temporal(train_gen_spatial, train_gen_temporal, fcnn_approximator, adam, metrics, shuffle=True, batch_size=100)
    assert train_epoch_loss > 0
    assert train_epoch_metrics['rmse'] > 0


def test__valid_2dspatial_temporal():
    DIFFUSIVITY = 0.3
    X_MIN, X_MAX = -1.0, 1.0
    Y_MIN, Y_MAX = -1.0, 1.0
    T_MIN, T_MAX = 0.0, 6.0

    def heat_equation_2d(u, x, y, t):
        left = diff(u, t) - DIFFUSIVITY * (diff(u, x, order=2) + diff(u, y, order=2))
        right = -torch.exp(-t) * ((X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN) - 2 * DIFFUSIVITY * (
                (Y_MAX - y) * (y - Y_MIN) + (X_MAX - x) * (x - X_MIN)))
        return left - right

    def analytical_solution(xx, yy, tt):
        return torch.exp(-tt) * (X_MAX - xx) * (xx - X_MIN) * (Y_MAX - yy) * (yy - Y_MIN)

    def rmse(uu, xx, yy, tt):
        error = uu - analytical_solution(xx, yy, tt)
        return torch.mean(error ** 2) ** 0.5

    metrics = {'rmse': rmse}

    def u0(x, y):
        return (X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN)

    initial_condition = FirstOrderInitialCondition(u0=u0)

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MIN, Y_MAX))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MAX, Y_MIN), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MAX), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MAX, Y_MIN))
    )

    fcnn = FCNN(
        n_input_units=3,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_2d,
        initial_condition=initial_condition,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )

    valid_gen_spatial = generator_2dspatial_rectangle(
        size=(8, 8), x_min=X_MIN, x_max=X_MAX, y_min=Y_MIN, y_max=Y_MAX
    )
    valid_gen_temporal = generator_temporal(size=8, t_min=T_MIN, t_max=T_MAX)

    valid_epoch_loss, valid_epoch_metrics = _valid_2dspatial_temporal(valid_gen_spatial, valid_gen_temporal, fcnn_approximator, metrics)
    assert valid_epoch_loss > 0
    assert valid_epoch_metrics['rmse'] > 0


def test__solve_2dspatial_temporal():
    DIFFUSIVITY = 0.3
    X_MIN, X_MAX = -1.0, 1.0
    Y_MIN, Y_MAX = -1.0, 1.0
    T_MIN, T_MAX = 0.0, 6.0

    def heat_equation_2d(u, x, y, t):
        left = diff(u, t) - DIFFUSIVITY * (diff(u, x, order=2) + diff(u, y, order=2))
        right = -torch.exp(-t) * ((X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN) - 2 * DIFFUSIVITY * (
                    (Y_MAX - y) * (y - Y_MIN) + (X_MAX - x) * (x - X_MIN)))
        return left - right

    def analytical_solution(xx, yy, tt):
        return torch.exp(-tt) * (X_MAX - xx) * (xx - X_MIN) * (Y_MAX - yy) * (yy - Y_MIN)

    def rmse(uu, xx, yy, tt):
        error = uu - analytical_solution(xx, yy, tt)
        return torch.mean(error ** 2) ** 0.5

    metrics = {'rmse': rmse}

    def u0(x, y):
        return (X_MAX - x) * (x - X_MIN) * (Y_MAX - y) * (y - Y_MIN)

    initial_condition = FirstOrderInitialCondition(u0=u0)

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MIN, Y_MAX))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MAX, Y_MIN), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MAX), end=(X_MAX, Y_MAX))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y, t: u,
        points_generator=generator_2dspatial_segment(size=16, start=(X_MIN, Y_MIN), end=(X_MAX, Y_MIN))
    )

    train_gen_spatial = generator_2dspatial_rectangle(
        size=(16, 16), x_min=X_MIN, x_max=X_MAX, y_min=Y_MIN, y_max=Y_MAX
    )
    train_gen_temporal = generator_temporal(size=16, t_min=T_MIN, t_max=T_MAX)
    valid_gen_spatial = generator_2dspatial_rectangle(
        size=(16, 16), x_min=X_MIN, x_max=X_MAX, y_min=Y_MIN, y_max=Y_MAX, random=False
    )
    valid_gen_temporal = generator_temporal(size=16, t_min=T_MIN, t_max=T_MAX, random=False)
    monitor = Monitor2DSpatialTemporal(
        check_on_x=torch.linspace(X_MIN, X_MAX, 32),
        check_on_y=torch.linspace(Y_MIN, Y_MAX, 32),
        check_on_t=torch.linspace(T_MIN, T_MAX, 4),
        check_every=10
    )

    fcnn = FCNN(
        n_input_units=3,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatialTemporal(
        single_network=fcnn,
        pde=heat_equation_2d,
        initial_condition=initial_condition,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )
    adam = optim.Adam(fcnn_approximator.parameters())

    heat_equation_2d_solution, _ = _solve_2dspatial_temporal(
        train_generator_spatial=train_gen_spatial,
        train_generator_temporal=train_gen_temporal,
        valid_generator_spatial=valid_gen_spatial,
        valid_generator_temporal=valid_gen_temporal,
        approximator=fcnn_approximator,
        optimizer=adam,
        batch_size=256,
        max_epochs=1,
        shuffle=True,
        metrics=metrics,
        monitor=monitor
    )

    xx, yy, tt = torch.rand(16), torch.rand(16), torch.rand(16)
    assert heat_equation_2d_solution(xx, yy, tt).shape == torch.Size([16])
    xx, yy, tt = torch.rand(16), torch.rand(16), torch.zeros(16)
    assert fcnn_approximator(xx, yy, tt).isclose(u0(xx, yy)).all()


def test__solve_2dspatial():
    def poisson_2d(u, xx, yy):
        return diff(u, xx, order=2) + diff(u, yy, order=2) - torch.sin(PI * xx) * torch.sin(PI * yy)

    def analytical_solution(xx, yy):
        return -1 / (2 * PI ** 2) * torch.sin(PI * xx) * torch.sin(PI * yy)

    metrics = {}

    def rmse(uu, xx, yy):
        error = uu - analytical_solution(xx, yy)
        return torch.mean(error ** 2) ** 0.5

    metrics['rmse'] = rmse

    dirichlet_boundary_left = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(0.0, 1.0))
    )
    dirichlet_boundary_right = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(1.0, 0.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_upper = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 1.0), end=(1.0, 1.0))
    )
    dirichlet_boundary_lower = BoundaryCondition(
        form=lambda u, x, y: u,
        points_generator=generator_2dspatial_segment(size=32, start=(0.0, 0.0), end=(1.0, 0.0))
    )

    fcnn = FCNN(
        n_input_units=2,
        n_output_units=1,
        hidden_units=(32, 32),
        actv=nn.Tanh,
    )
    fcnn_approximator = SingleNetworkApproximator2DSpatial(
        single_network=fcnn,
        pde=poisson_2d,
        boundary_conditions=[
            dirichlet_boundary_left,
            dirichlet_boundary_right,
            dirichlet_boundary_upper,
            dirichlet_boundary_lower
        ]
    )
    adam = optim.Adam(fcnn_approximator.parameters(), lr=0.0005)

    train_gen_spatial = generator_2dspatial_rectangle(size=(32, 32), x_min=0.0, x_max=1.0, y_min=0.0, y_max=1.0)
    valid_gen_spatial = generator_2dspatial_rectangle(size=(20, 20), x_min=0.0, x_max=1.0, y_min=0.0, y_max=1.0,
                                                      random=False)

    poisson_2d_solution, _ = _solve_2dspatial(
        train_generator_spatial=train_gen_spatial,
        valid_generator_spatial=valid_gen_spatial,
        approximator=fcnn_approximator,
        optimizer=adam,
        batch_size=256,
        max_epochs=1,
        shuffle=True,
        metrics=metrics,
        monitor=Monitor2DSpatial(
            check_on_x=torch.linspace(0.0, 1.0, 20),
            check_on_y=torch.linspace(0.0, 1.0, 20),
            check_every=100
        )
    )
    xx, yy = torch.rand(16), torch.rand(16)
    assert poisson_2d_solution(xx, yy).shape == torch.Size([16])



import torch
import pytest
from neurodiffeq.operators import grad

N_SAMPLES = 10

f1 = lambda x, y, z: x ** 2 + y ** 2 + z ** 2
g1 = lambda x, y, z: [2 * x, 2 * y, 2 * z]
f2 = lambda x, y, z: x ** 2 + y ** 2
g2 = lambda x, y, z: [2 * x, 2 * y, torch.zeros_like(z, requires_grad=z.requires_grad)]
f3 = lambda x, y, z: torch.zeros_like(x, requires_grad=True)
g3 = lambda x, y, z: list(map(lambda v: torch.zeros_like(v, requires_grad=v.requires_grad), [x, y, z]))


@pytest.fixture
def x3():
    return [torch.rand((N_SAMPLES, 1), requires_grad=True) for _ in range(3)]


@pytest.mark.parametrize(
    argnames=['f', 'g_func'],
    argvalues=list(zip([f1, f2, f3], [g1, g2, g3])),
)
def test_grad(x3, f, g_func):
    g3_true = g_func(*x3)
    g3_pred = grad(f(*x3), *x3)

    for x, g_true, g_pred in zip(x3, g3_true, g3_pred):
        assert isinstance(g_pred, torch.Tensor)
        assert g_pred.requires_grad
        assert torch.isclose(g_true, g_pred).all()



import pytest
import torch
import torch.nn as nn
from torch import sin, cos
import numpy as np
from neurodiffeq.generators import GeneratorSpherical
from neurodiffeq.function_basis import ZonalSphericalHarmonics
from neurodiffeq.networks import FCNN
from neurodiffeq.operators import spherical_curl
from neurodiffeq.operators import spherical_grad
from neurodiffeq.operators import spherical_div
from neurodiffeq.operators import spherical_laplacian
from neurodiffeq.operators import spherical_vector_laplacian
from neurodiffeq.operators import grad


@pytest.fixture(autouse=True)
def magic():
    torch.manual_seed(42)
    np.random.seed(42)


class HarmonicsNN(nn.Module):
    def __init__(self, degrees, harmonics_fn):
        super(HarmonicsNN, self).__init__()
        self.net_r = FCNN(1, n_output_units=len(degrees))
        self.harmonics_fn = harmonics_fn

    def forward(self, r, theta, phi):
        R = self.net_r(r)
        Y = self.harmonics_fn(theta, phi)
        return (R * Y).sum(dim=1, keepdim=True)


EPS = 1e-4
degrees = list(range(10))


@pytest.fixture
def x():
    n_points, r_min, r_max = 1024, 1.0, 10.0
    g = GeneratorSpherical(n_points, r_min=r_min, r_max=r_max)
    return [t.reshape(-1, 1) for t in g.get_examples()]


@pytest.fixture
def U(x):
    F = [HarmonicsNN(degrees, ZonalSphericalHarmonics(degrees=degrees)) for _ in range(3)]
    return list(map(lambda f: f(*x), F))


@pytest.fixture
def u(x):
    return HarmonicsNN(degrees, ZonalSphericalHarmonics(degrees=degrees))(*x)


def is_zero(t):
    if isinstance(t, (tuple, list)):
        for i in t:
            if not is_zero(i):
                return False
        return True
    elif isinstance(t, torch.Tensor):
        return t.detach().cpu().max() < EPS
    else:
        raise ValueError(f"t must be list, tuple or tensor; got {type(t)}")


def test_div_curl(U, x):
    curl_u = spherical_curl(*U, *x)
    div_curl_u = spherical_div(*curl_u, *x)
    assert is_zero(div_curl_u), div_curl_u


def test_curl_grad(u, x):
    grad_u = spherical_grad(u, *x)
    curl_grad_u = spherical_curl(*grad_u, *x)
    assert is_zero(curl_grad_u), curl_grad_u


def test_div_grad(u, x):
    grad_u = spherical_grad(u, *x)
    div_grad_u = spherical_div(*grad_u, *x)
    lap_u = spherical_laplacian(u, *x)
    delta = div_grad_u - lap_u
    assert is_zero(delta), delta


def test_laplacian(u, x):
    test_div_grad(u, x)


def test_curl_curl(U, x):
    curl_curl_u = spherical_curl(*spherical_curl(*U, *x), *x)
    grad_div_u = spherical_grad(spherical_div(*U, *x), *x)
    vec_lap_u = spherical_vector_laplacian(*U, *x)

    vec_delta = [cc - (gd - vl) for cc, gd, vl in zip(curl_curl_u, grad_div_u, vec_lap_u)]
    assert is_zero(vec_delta), vec_delta


def test_vec_laplacian(U, x):
    test_curl_curl(U, x)



import numpy as np
from numpy import isclose
import pytest
from pytest import raises, warns
from scipy.integrate import odeint
import matplotlib

matplotlib.use('Agg')  # use a non-GUI backend, so plots are not shown during testing

from neurodiffeq.neurodiffeq import safe_diff as diff
from neurodiffeq.networks import FCNN, SinActv
from neurodiffeq.ode import IVP, DirichletBVP
from neurodiffeq.ode import solve, solve_system, Monitor
from neurodiffeq.monitors import Monitor1D
from neurodiffeq.solvers import Solution1D, Solver1D
from neurodiffeq.generators import Generator1D, BaseGenerator

import torch


@pytest.fixture(autouse=True)
def magic():
    MAGIC = 42
    torch.manual_seed(MAGIC)
    np.random.seed(MAGIC)


def test_monitor():
    exponential = lambda u, t: diff(u, t) - u
    init_val_ex = IVP(t_0=0.0, u_0=1.0)
    with warns(FutureWarning):
        solution_ex, _ = solve(ode=exponential, condition=init_val_ex,
                               t_min=0.0, t_max=2.0,
                               max_epochs=3,
                               monitor=Monitor1D(t_min=0.0, t_max=2.0, check_every=1))

    with warns(FutureWarning):
        Monitor(t_min=0.0, t_max=2.0, check_every=1)


def test_ode():
    def mse(u, t):
        true_u = torch.sin(t)
        return torch.mean((u - true_u) ** 2)

    exponential = lambda u, t: diff(u, t) - u
    init_val_ex = IVP(t_0=0.0, u_0=1.0)
    with warns(FutureWarning):
        solution_ex, loss_history = solve(ode=exponential, condition=init_val_ex,
                                          t_min=0.0, t_max=2.0, shuffle=False,
                                          max_epochs=10, return_best=True, metrics={'mse': mse})

    assert isinstance(solution_ex, Solution1D)
    assert isinstance(loss_history, dict)
    keys = ['train_loss', 'valid_loss']
    for key in keys:
        assert key in loss_history
        assert isinstance(loss_history[key], list)
    assert len(loss_history[keys[0]]) == len(loss_history[keys[1]])


def test_ode_system():
    parametric_circle = lambda u1, u2, t: [diff(u1, t) - u2, diff(u2, t) + u1]
    init_vals_pc = [IVP(t_0=0.0, u_0=0.0), IVP(t_0=0.0, u_0=1.0)]

    with warns(FutureWarning):
        solution_pc, loss_history = solve_system(
            ode_system=parametric_circle,
            conditions=init_vals_pc,
            t_min=0.0,
            t_max=2 * np.pi,
            max_epochs=10,
        )

    assert isinstance(solution_pc, Solution1D)
    assert isinstance(loss_history, dict)
    keys = ['train_loss', 'valid_loss']
    for key in keys:
        assert key in loss_history
        assert isinstance(loss_history[key], list)
    assert len(loss_history[keys[0]]) == len(loss_history[keys[1]])


def test_additional_loss_term():
    def particle_squarewell(y1, y2, t):
        return [
            (-1 / 2) * diff(y1, t, order=2) - 3 - (y2) * (y1),
            diff(y2, t)
        ]

    def zero_y2(y1, y2, t):
        return torch.sum(y2 ** 2)

    boundary_conditions = [
        DirichletBVP(t_0=0, u_0=0, t_1=2, u_1=0),
        DirichletBVP(t_0=0, u_0=0, t_1=2, u_1=0),
    ]

    with warns(FutureWarning):
        solution_squarewell, loss_history = solve_system(
            ode_system=particle_squarewell, conditions=boundary_conditions,
            additional_loss_term=zero_y2,
            t_min=0.0, t_max=2.0,
            max_epochs=10,
        )
    assert isinstance(solution_squarewell, Solution1D)
    assert isinstance(loss_history, dict)
    keys = ['train_loss', 'valid_loss']
    for key in keys:
        assert key in loss_history
        assert isinstance(loss_history[key], list)
    assert len(loss_history[keys[0]]) == len(loss_history[keys[1]])


def test_solution():
    t0, u0 = np.random.rand() + 0, np.random.rand() + 0
    t1, u1 = np.random.rand() + 1, np.random.rand() + 1
    N_SAMPLES = 100

    def get_solution(use_single: bool) -> Solution1D:
        conditions = [IVP(t0, u0), IVP(t1, u1)]
        if use_single:
            net = FCNN(1, 2)
            with warns(DeprecationWarning):
                for i, cond in enumerate(conditions):
                    cond.set_impose_on(i)
            return Solution1D(net, conditions)
        else:
            nets = [FCNN(1, 1), FCNN(1, 1)]
            return Solution1D(nets, conditions)

    def check_output(us, shape, type, msg=""):
        msg += " "
        assert isinstance(us, (list, tuple)), msg + "returned type is not a list"
        assert len(us) == 2, msg + "returned length is not 2"
        assert isinstance(us[0], type) and isinstance(us[1], type), msg + f"returned element is not {type}"
        assert us[0].shape == shape and us[1].shape == shape, msg + f"returned element shape is not {shape}"
        assert us[0][0] == u0, msg + f"first condition is not properly imposed"
        assert us[1][-1] == u1, msg + f"second condition is not properly imposed"

    for use_single in [True, False]:
        solution = get_solution(use_single=use_single)
        ts = torch.linspace(t0, t1, N_SAMPLES)
        us = solution(ts)
        check_output(us, shape=(N_SAMPLES,), type=torch.Tensor, msg=f"[use_single={use_single}]")
        with warns(FutureWarning):
            us = solution(ts, as_type='np')
        check_output(us, shape=(N_SAMPLES,), type=np.ndarray, msg=f"[use_single={use_single}]")
        us = solution(ts, to_numpy=True)
        check_output(us, shape=(N_SAMPLES,), type=np.ndarray, msg=f"[use_single={use_single}]")

        ts = ts.reshape(-1, 1)
        us = solution(ts)
        check_output(us, shape=(N_SAMPLES, 1), type=torch.Tensor, msg=f"[use_single={use_single}]")
        with warns(FutureWarning):
            us = solution(ts, as_type='np')
        check_output(us, shape=(N_SAMPLES, 1), type=np.ndarray, msg=f"[use_single={use_single}]")
        us = solution(ts, to_numpy=True)
        check_output(us, shape=(N_SAMPLES, 1), type=np.ndarray, msg=f"[use_single={use_single}]")


def test_get_internals():
    parametric_circle = lambda x1, x2, t: [diff(x1, t) - x2, diff(x2, t) + x1]

    init_vals_pc = [
        IVP(t_0=0.0, u_0=0.0),
        IVP(t_0=0.0, u_0=1.0),
    ]

    solver = Solver1D(
        ode_system=parametric_circle,
        conditions=init_vals_pc,
        t_min=0.0,
        t_max=2 * np.pi,
    )

    solver.fit(max_epochs=1)
    internals = solver.get_internals()
    assert isinstance(internals, dict)
    internals = solver.get_internals(return_type='list')
    assert isinstance(internals, dict)
    internals = solver.get_internals(return_type='dict')
    assert isinstance(internals, dict)
    internals = solver.get_internals(['generator', 'n_batches'], return_type='dict')
    assert isinstance(internals, dict)
    internals = solver.get_internals(['generator', 'n_batches'], return_type='list')
    assert isinstance(internals, list)
    internals = solver.get_internals('train_generator')
    assert isinstance(internals, BaseGenerator)



import torch
import numpy as np
import torch.nn as nn
import pytest
from neurodiffeq.networks import FCNN
from neurodiffeq.networks import Resnet
from neurodiffeq.networks import MonomialNN
from neurodiffeq.networks import SinActv
from neurodiffeq.networks import Swish

MAGIC = 42
torch.manual_seed(MAGIC)
np.random.seed(MAGIC)

N_TESTS = 5


def _test_shape(n_samples, n_features_in, n_features_out, model_constructor, *args, **kwargs):
    net = model_constructor(*args, **kwargs)
    x = torch.rand(n_samples, n_features_in).requires_grad_(True)
    y = net(x)
    assert y.shape == (n_samples, n_features_out)


def test_fcnn():
    for _ in range(N_TESTS):
        n_samples = np.random.randint(30, 100)
        n_features_in = np.random.randint(1, 5)
        n_features_out = np.random.randint(1, 5)
        n_hidden_units = np.random.randint(30, 60)
        n_hidden_layers = np.random.randint(0, 4)
        hidden_units = [np.random.randint(1, 10) for _ in range(np.random.randint(2, 4))]

        with pytest.warns(FutureWarning):
            _test_shape(
                n_samples, n_features_in, n_features_out, FCNN,
                n_input_units=n_features_in,
                n_output_units=n_features_out,
                n_hidden_units=n_hidden_units,
            )
        with pytest.warns(FutureWarning):
            _test_shape(
                n_samples, n_features_in, n_features_out, FCNN,
                n_input_units=n_features_in,
                n_output_units=n_features_out,
                n_hidden_layers=n_hidden_layers,
            )
        with pytest.warns(FutureWarning):
            _test_shape(
                n_samples, n_features_in, n_features_out, FCNN,
                n_input_units=n_features_in,
                n_output_units=n_features_out,
                n_hidden_units=n_hidden_units,
                n_hidden_layers=n_hidden_layers,
            )
        with pytest.warns(FutureWarning):
            _test_shape(
                n_samples, n_features_in, n_features_out, FCNN,
                n_input_units=n_features_in,
                n_output_units=n_features_out,
                n_hidden_units=n_hidden_units,
                n_hidden_layers=n_hidden_layers,
                hidden_units=hidden_units
            )

        _test_shape(
            n_samples, n_features_in, n_features_out, FCNN,
            n_input_units=n_features_in,
            n_output_units=n_features_out,
            hidden_units=hidden_units
        )


def test_resnet():
    for _ in range(N_TESTS):
        n_samples = np.random.randint(30, 100)
        n_features_in = np.random.randint(1, 5)
        n_features_out = np.random.randint(1, 5)
        n_hidden_units = np.random.randint(30, 60)
        n_hidden_layers = np.random.randint(0, 4)
        hidden_units = [np.random.randint(1, 10) for _ in range(np.random.randint(2, 4))]
        with pytest.warns(FutureWarning):
            _test_shape(
                n_samples, n_features_in, n_features_out, Resnet,
                n_input_units=n_features_in,
                n_output_units=n_features_out,
                n_hidden_units=n_hidden_units,
            )
        with pytest.warns(FutureWarning):
            _test_shape(
                n_samples, n_features_in, n_features_out, Resnet,
                n_input_units=n_features_in,
                n_output_units=n_features_out,
                n_hidden_layers=n_hidden_layers,
            )
        with pytest.warns(FutureWarning):
            _test_shape(
                n_samples, n_features_in, n_features_out, Resnet,
                n_input_units=n_features_in,
                n_output_units=n_features_out,
                n_hidden_units=n_hidden_units,
                n_hidden_layers=n_hidden_layers,
            )
        with pytest.warns(FutureWarning):
            _test_shape(
                n_samples, n_features_in, n_features_out, Resnet,
                n_input_units=n_features_in,
                n_output_units=n_features_out,
                n_hidden_units=n_hidden_units,
                n_hidden_layers=n_hidden_layers,
                hidden_units=hidden_units
            )

        _test_shape(
            n_samples, n_features_in, n_features_out, Resnet,
            n_input_units=n_features_in,
            n_output_units=n_features_out,
            hidden_units=hidden_units
        )


def test_monomial_nn():
    ALL_DEGREES = list(range(1, N_TESTS + 1))
    ALL_DEGREES += [-d for d in ALL_DEGREES]

    for test_id in range(N_TESTS):
        degrees = np.random.choice(ALL_DEGREES, size=test_id + 1, replace=False)
        n_samples = np.random.randint(30, 100)
        n_features_in = np.random.randint(1, 5)
        net = MonomialNN(degrees)
        x = torch.rand(n_samples, n_features_in) + 0.5
        y = net(x)
        for i, d in enumerate(degrees):
            x_d = y[:, i * n_features_in: (i + 1) * n_features_in]
            assert (x_d - x ** d).abs().max() < 1e-3


def test_swish():
    x = torch.rand(10, 5)

    f = Swish()
    assert len(list(f.parameters())) == 0
    assert torch.isclose(f(x), x * torch.sigmoid(x)).all()

    beta = 3.0
    f = Swish(beta, trainable=True)
    assert len(list(f.parameters())) == 1
    assert list(f.parameters())[0].shape == ()
    assert torch.isclose(f(x), x * torch.sigmoid(beta * x)).all()



import torch
import numpy as np
import random
import pytest
from pytest import raises, warns, deprecated_call
from neurodiffeq.conditions import NoCondition
from neurodiffeq.conditions import IVP
from neurodiffeq.conditions import BundleIVP
from neurodiffeq.conditions import EnsembleCondition
from neurodiffeq.conditions import DirichletBVP
from neurodiffeq.conditions import DirichletBVP2D
from neurodiffeq.conditions import DirichletBVPSpherical
from neurodiffeq.conditions import InfDirichletBVPSpherical
from neurodiffeq.conditions import DirichletBVPSphericalBasis
from neurodiffeq.conditions import InfDirichletBVPSphericalBasis
from neurodiffeq.conditions import IBVP1D
from neurodiffeq.conditions import DoubleEndedBVP1D
from neurodiffeq.networks import FCNN
from neurodiffeq.neurodiffeq import safe_diff as diff


@pytest.fixture(autouse=True)
def magic():
    MAGIC = 42
    torch.manual_seed(MAGIC)
    np.random.seed(MAGIC)
    return MAGIC


@pytest.fixture
def N_SAMPLES():
    return 10


@pytest.fixture
def ones(N_SAMPLES):
    return torch.ones(N_SAMPLES, 1, requires_grad=True)


@pytest.fixture
def lin(N_SAMPLES):
    return torch.linspace(0, 1, N_SAMPLES, requires_grad=True).reshape(-1, 1)


@pytest.fixture
def x0():
    return random.random()


@pytest.fixture
def x1():
    return random.random()


@pytest.fixture
def y0():
    return random.random()


@pytest.fixture
def y1():
    return random.random()


@pytest.fixture
def t0():
    return random.random()


@pytest.fixture
def t1():
    return random.random()


@pytest.fixture
def u00():
    return random.random()


@pytest.fixture
def u01():
    return random.random()


@pytest.fixture
def u10():
    return random.random()


@pytest.fixture
def u11():
    return random.random()


@pytest.fixture
def net11():
    return FCNN(1, 1)


@pytest.fixture
def net21():
    return FCNN(2, 1)


@pytest.fixture
def net12():
    return FCNN(1, 2)


@pytest.fixture
def net31():
    return FCNN(3, 1)


@pytest.fixture
def net41():
    return FCNN(4, 1)


@pytest.fixture
def boundary_functions_2d(x0, x1, y0, y1, u00, u01, u10, u11):
    net_f0, net_f1, net_g0, net_g1 = [FCNN(1, 1) for _ in range(4)]
    cond_f0 = DirichletBVP(y0, u00, y1, u01)
    cond_f1 = DirichletBVP(y0, u10, y1, u11)
    cond_g0 = DirichletBVP(x0, u00, x1, u10)
    cond_g1 = DirichletBVP(x0, u01, x1, u11)
    f0 = lambda y: cond_f0.enforce(net_f0, y)
    f1 = lambda y: cond_f1.enforce(net_f1, y)
    g0 = lambda x: cond_g0.enforce(net_g0, x)
    g1 = lambda x: cond_g1.enforce(net_g1, x)

    return f0, f1, g0, g1


def all_close(x_tensor, y_tensor, rtol=5e-4, atol=1e-6, equal_nan=False):
    if isinstance(y_tensor, (float, int)):
        y_tensor = torch.ones_like(x_tensor) * y_tensor
    return torch.isclose(x_tensor, y_tensor, rtol=rtol, atol=atol, equal_nan=equal_nan).all()


def test_no_condition(N_SAMPLES):
    N_INPUTS = 5
    N_OUTPUTS = 5

    for n_in, n_out in zip(range(1, N_INPUTS), range(1, N_OUTPUTS)):
        xs = [torch.rand(N_SAMPLES, 1, requires_grad=True) for _ in range(n_in)]
        net = FCNN(n_in, n_out)

        cond = NoCondition()
        y_cond = cond.enforce(net, *xs)
        y_raw = net(torch.cat(xs, dim=1))
        assert (y_cond == y_raw).all()


def test_ivp(x0, y0, y1, ones, net11):
    x = x0 * ones
    cond = IVP(x0, y0)
    y = cond.enforce(net11, x)
    assert torch.isclose(y, y0 * ones).all(), "y(x_0) != y_0"

    cond = IVP(x0, y0, y1)
    y = cond.enforce(net11, x)
    assert all_close(y, y0), "y(x_0) != y_0"
    assert all_close(diff(y, x), y1), "y'(x_0) != y'_0"


def test_bundleivp(x0, y0, y1, ones, lin, net11, net21, net31, net41):
    # Regular IVP with no bundle:
    x = x0 * ones
    cond = BundleIVP(x0, y0)
    y = cond.enforce(net11, x)
    assert torch.isclose(y, y0 * ones).all(), "y(x_0) != y_0"

    cond = BundleIVP(x0, y0, y1)
    y = cond.enforce(net11, x)
    assert all_close(y, y0), "y(x_0) != y_0"
    assert all_close(diff(y, x), y1), "y'(x_0) != y'_0"

    # Bundle in u_0:
    y_bundle = y0 * lin
    cond = BundleIVP(t_0=x0, bundle_conditions=['u_0'])
    y = cond.enforce(net21, x, y_bundle)
    assert torch.isclose(y, y0 * lin).all(), "y(x_0) != y_0"

    cond = BundleIVP(t_0=x0, u_0_prime=y1, bundle_conditions=['u_0'])
    y = cond.enforce(net21, x, y_bundle)
    assert torch.isclose(y, y0 * lin).all(), "y(x_0) != y_0"
    assert all_close(diff(y, x), y1), "y'(x_0) != y'_0"

    # Bundle in u_0_prime:
    y_prime_bundle = y1 * lin
    cond = BundleIVP(t_0=x0, u_0=y0, bundle_conditions=['u_0_prime'])
    y = cond.enforce(net21, x, y_prime_bundle)
    assert all_close(y, y0), "y(x_0) != y_0"
    assert torch.isclose(diff(y, x), y1 * lin).all(), "y'(x_0) != y'_0"

    # Bundle in u_0 and u_0_prime:
    cond = BundleIVP(t_0=x0, bundle_conditions=['u_0', 'u_0_prime'])
    y = cond.enforce(net31, x, y_bundle, y_prime_bundle)
    assert torch.isclose(y, y0 * lin).all(), "y(x_0) != y_0"
    assert torch.isclose(diff(y, x), y1 * lin).all(), "y'(x_0) != y'_0"

    # Bundle in t_0:
    x = x0 * lin
    x_bundle = x0 * lin
    cond = BundleIVP(u_0=y0, bundle_conditions=['t_0'])
    y = cond.enforce(net21, x, x_bundle)
    assert torch.isclose(y, y0 * ones).all(), "y(x_0) != y_0"

    cond = BundleIVP(u_0=y0, u_0_prime=y1, bundle_conditions=['t_0'])
    y = cond.enforce(net21, x, x_bundle)
    assert all_close(y, y0), "y(x_0) != y_0"
    assert all_close(diff(y, x), y1), "y'(x_0) != y'_0"

    # Bundle in t_0 and u_0:
    cond = BundleIVP(bundle_conditions=['t_0', 'u_0'])
    y = cond.enforce(net31, x, x_bundle, y_bundle)
    assert torch.isclose(y, y0 * lin).all(), "y(x_0) != y_0"

    cond = BundleIVP(u_0_prime=y1, bundle_conditions=['t_0', 'u_0'])
    y = cond.enforce(net31, x, x_bundle, y_bundle)
    assert torch.isclose(y, y0 * lin).all(), "y(x_0) != y_0"
    assert all_close(diff(y, x), y1), "y'(x_0) != y'_0"

    # Bundle in t_0, u_0 and u_0_prime:
    cond = BundleIVP(bundle_conditions=['t_0', 'u_0', 'u_0_prime'])
    y = cond.enforce(net41, x, x_bundle, y_bundle, y_prime_bundle)
    assert torch.isclose(y, y0 * lin).all(), "y(x_0) != y_0"
    assert torch.isclose(diff(y, x), y1 * lin).all(), "y'(x_0) != y'_0"


def test_ivp_legacy_signature():
    with warns(FutureWarning):
        IVP(0, x_0=1)
    with warns(FutureWarning):
        IVP(0, 1, x_0_prime=2)
    with warns(FutureWarning):
        IVP(0, x_0=1, x_0_prime=2)
    with raises(KeyError):
        IVP(0, x_0=1, u_0=2)
    with raises(KeyError):
        IVP(0, x_0_prime=1, u_0_prime=2)


def test_ensemble_condition(x0, x1, y0, y1, ones, net12):
    cond = EnsembleCondition(
        IVP(x0, y0),
        IVP(x1, y0, y1),
    )

    x = x0 * ones
    y = cond.enforce(net12, x)
    ya = y[:, 0:1]
    assert all_close(ya, y0), "y(x_0) != y_0"
    x = x1 * ones
    y = cond.enforce(net12, x)
    yb = y[:, 1:2]
    assert all_close(yb, y0), "y(x_0) != y_0"
    assert all_close(diff(yb, x), y1), "y'(x_0) != y'_0"

    net12 = FCNN(1, 1)
    cond = EnsembleCondition(
        IVP(x0, y0),
    )
    x = x0 * ones
    y = cond.enforce(net12, x)
    assert all_close(y, y0), "y(x_0) != y_0"


def test_dirichlet_bvp(x0, x1, y0, y1, ones, net11):
    cond = DirichletBVP(x0, y0, x1, y1)

    x = x0 * ones
    y = cond.enforce(net11, x)
    assert all_close(y, y0), "y(x_0) != y_0"

    x = x1 * ones
    y = cond.enforce(net11, x)
    assert all_close(y, y1), "y(x_1) != y_1"


def test_bvp_legacy_signature(x0, x1, y0, y1):
    with warns(FutureWarning):
        DirichletBVP(t_0=0, t_1=0, x_0=0, x_1=0)
    with warns(FutureWarning):
        DirichletBVP(t_0=0, x_0=0, t_1=0, x_1=0)
    with warns(FutureWarning):
        DirichletBVP(0, 2, t_1=0, x_1=0)
    with warns(FutureWarning):
        DirichletBVP(t_0=0, x_0=0, t_1=0, u_1=0)
    with warns(FutureWarning):
        DirichletBVP(t_0=0, u_0=0, t_1=0, x_1=0)
    with raises(KeyError), warns(FutureWarning):
        DirichletBVP(t_0=0, u_0=0, x_0=0, t_1=0, x_1=0)
    with raises(KeyError), warns(FutureWarning):
        DirichletBVP(t_0=0, x_0=0, t_1=0, x_1=0, u_1=0)
    with raises(KeyError), warns(FutureWarning):
        DirichletBVP(t_0=0, u_0=0, x_0=0, t_1=0, x_1=0, u_1=0)


def test_dirichlet_bvp_2d(x0, x1, y0, y1, u00, u01, u10, u11, ones, net21, boundary_functions_2d):
    # set the boundary conditions on the four sides
    f0, f1, g0, g1 = boundary_functions_2d

    # test whether condition is enforced
    condition = DirichletBVP2D(x0, f0, x1, f1, y0, g0, y1, g1)

    x = x0 * ones
    y = torch.linspace(y0, y1, ones.numel(), requires_grad=True).reshape(-1, 1)
    assert all_close(condition.enforce(net21, x, y), f0(y)), "left boundary not satisfied"

    x = x1 * ones
    y = torch.linspace(y0, y1, ones.numel(), requires_grad=True).reshape(-1, 1)
    assert all_close(condition.enforce(net21, x, y), f1(y)), "right boundary not satisfied"

    x = torch.linspace(x0, x1, ones.numel(), requires_grad=True).reshape(-1, 1)
    y = y0 * ones
    assert all_close(condition.enforce(net21, x, y), g0(x)), "lower boundary not satisfied"

    x = torch.linspace(x0, x1, ones.numel(), requires_grad=True).reshape(-1, 1)
    y = y1 * ones
    assert all_close(condition.enforce(net21, x, y), g1(x)), "upper boundary not satisfied"


def test_ibvp_1d(x0, x1, t0, t1, u00, u01, u10, u11, ones, net21, N_SAMPLES):
    # set the initial condition ut0(x) = u(x, t0)
    net_ut0 = FCNN(1, 1)
    cond_ut0 = DirichletBVP(x0, u00, x1, u10)
    ut0 = lambda x: cond_ut0.enforce(net_ut0, x)

    # set the Dirichlet boundary conditions g(t) = u(x0, t) and h(t) = u(x1, t)
    net_g, net_h = FCNN(1, 1), FCNN(1, 1)
    cond_g = IVP(t0, u00)
    cond_h = IVP(t0, u10)
    g = lambda t: cond_g.enforce(net_g, t)
    h = lambda t: cond_h.enforce(net_h, t)

    # set the Neumann boundary conditions p(t) = u'_x(x0, t) and q(t) = u'_x(x1, t)
    x = x0 * ones
    p0 = diff(ut0(x), x)[0, 0].item()
    x = x1 * ones
    q0 = diff(ut0(x), x)[0, 0].item()
    p1, q1 = random.random(), random.random()
    net_p, net_q = FCNN(1, 1), FCNN(1, 1)
    cond_p = DirichletBVP(t0, p0, t1, p1)
    cond_q = DirichletBVP(t0, q0, t1, q1)
    p = lambda t: cond_p.enforce(net_p, t)
    q = lambda t: cond_q.enforce(net_q, t)

    # test Dirichlet-Dirichlet condition
    condition = IBVP1D(x0, x1, t0, ut0, x_min_val=g, x_max_val=h)
    x = torch.linspace(x0, x1, N_SAMPLES, requires_grad=True).view(-1, 1)
    t = t0 * ones
    assert all_close(condition.enforce(net21, x, t), ut0(x)), "initial condition not satisfied"
    x = x0 * ones
    t = torch.linspace(t0, t1, N_SAMPLES, requires_grad=True).view(-1, 1)
    assert all_close(condition.enforce(net21, x, t), g(t)), "left Dirichlet BC not satisfied"
    x = x1 * ones
    t = torch.linspace(t0, t1, N_SAMPLES, requires_grad=True).view(-1, 1)
    assert all_close(condition.enforce(net21, x, t), h(t)), "right Dirichlet BC not satisfied"

    # test Dirichlet-Neumann condition
    condition = IBVP1D(x0, x1, t0, ut0, x_min_val=g, x_max_prime=q)
    x = torch.linspace(x0, x1, N_SAMPLES, requires_grad=True).view(-1, 1)
    t = t0 * ones
    assert all_close(condition.enforce(net21, x, t), ut0(x)), "initial condition not satisfied"
    x = x0 * ones
    t = torch.linspace(t0, t1, N_SAMPLES, requires_grad=True).view(-1, 1)
    assert all_close(condition.enforce(net21, x, t), g(t)), "left Dirichlet BC not satisfied"
    x = x1 * ones
    t = torch.linspace(t0, t1, N_SAMPLES, requires_grad=True).view(-1, 1)
    assert all_close(diff(condition.enforce(net21, x, t), x), q(t)), "right Neumann BC not satisfied"

    # test Neumann-Dirichlet condition
    condition = IBVP1D(x0, x1, t0, ut0, x_min_prime=p, x_max_val=h)
    x = torch.linspace(x0, x1, N_SAMPLES, requires_grad=True).view(-1, 1)
    t = t0 * ones
    assert all_close(condition.enforce(net21, x, t), ut0(x)), "initial condition not satisfied"
    x = x0 * ones
    t = torch.linspace(t0, t1, N_SAMPLES, requires_grad=True).view(-1, 1)
    assert all_close(diff(condition.enforce(net21, x, t), x), p(t)), "left Neumann BC not satisfied"
    x = x1 * ones
    t = torch.linspace(t0, t1, N_SAMPLES, requires_grad=True).view(-1, 1)
    assert all_close(condition.enforce(net21, x, t), h(t)), "right Dirichlet BC not satisfied"

    # test Neumann-Neumann condition
    condition = IBVP1D(x0, x1, t0, ut0, x_min_prime=p, x_max_prime=q)
    x = torch.linspace(x0, x1, N_SAMPLES, requires_grad=True).view(-1, 1)
    t = t0 * ones
    assert all_close(condition.enforce(net21, x, t), ut0(x)), "initial condition not satisfied"
    x = x0 * ones
    t = torch.linspace(t0, t1, N_SAMPLES, requires_grad=True).view(-1, 1)
    assert all_close(diff(condition.enforce(net21, x, t), x), p(t)), "left Neumann BC not satisfied"
    x = x1 * ones
    t = torch.linspace(t0, t1, N_SAMPLES, requires_grad=True).view(-1, 1)
    assert all_close(diff(condition.enforce(net21, x, t), x), q(t)), "right Neumann BC not satisfied"

    # test unimplemented combination of conditions
    with raises(NotImplementedError):
        IBVP1D(
            t_min=0, t_min_val=lambda x: 0,
            x_min=0, x_min_val=None, x_min_prime=None,
            x_max=1, x_max_val=None, x_max_prime=None,
        )
    with raises(NotImplementedError):
        IBVP1D(
            t_min=0, t_min_val=lambda x: 0,
            x_min=0, x_min_val=lambda t: 0, x_min_prime=lambda t: 0,
            x_max=1, x_max_val=None, x_max_prime=None,
        )
    with raises(NotImplementedError):
        IBVP1D(
            t_min=0, t_min_val=lambda x: 0,
            x_min=0, x_min_val=None, x_min_prime=lambda t: 0,
            x_max=1, x_max_val=None, x_max_prime=None,
        )


def test_dirichlet_bvp_spherical(x0, x1, ones, net31, N_SAMPLES):
    r0, r1 = x0, x1
    r2 = (r0 + r1) / 2

    no_condition = NoCondition()
    # B.C. for the interior boundary (r_min)
    net_f = FCNN(2, 1)
    f = lambda th, ph: no_condition.enforce(net_f, th, ph)

    # B.C. for the exterior boundary (r_max)
    net_g = FCNN(2, 1)
    g = lambda th, ph: no_condition.enforce(net_g, th, ph)

    condition = DirichletBVPSpherical(r_0=r0, f=f, r_1=r1, g=g)

    theta = torch.rand(N_SAMPLES, 1) * np.pi
    phi = torch.rand(N_SAMPLES, 1) * 2 * np.pi
    r = r0 * ones
    assert all_close(condition.enforce(net31, r, theta, phi), f(theta, phi)), "inner Dirichlet BC not satisfied"
    r = r1 * ones
    assert all_close(condition.enforce(net31, r, theta, phi), g(theta, phi)), "inner Dirichlet BC not satisfied"

    condition = DirichletBVPSpherical(r_0=r2, f=f)
    r = r2 * ones
    assert all_close(condition.enforce(net31, r, theta, phi), f(theta, phi)), "single ended BC not satisfied"


def test_inf_dirichlet_bvp_spherical(ones, net31):
    r0 = random.random()
    r1 = 1e15
    no_condition = NoCondition()
    net_f, net_g = FCNN(2, 1), FCNN(2, 1)

    # B.C. for the interior boundary (r=r_min)
    f = lambda th, ph: no_condition.enforce(net_f, th, ph)
    # B.C. for the exterior boundary (r=infinity)
    g = lambda th, ph: no_condition.enforce(net_g, th, ph)

    condition = InfDirichletBVPSpherical(r_0=r0, f=f, g=g, order=1)
    theta = torch.rand(10, 1) * np.pi
    phi = torch.rand(10, 1) * (2 * np.pi)

    r = r0 * ones
    assert all_close(condition.enforce(net31, r, theta, phi), f(theta, phi)), "inner DirichletBC not satisfied"
    r = r1 * ones
    assert all_close(condition.enforce(net31, r, theta, phi), g(theta, phi)), "Infinity DirichletBC not satisfied"


def test_dirichlet_bvp_spherical_basis(x0, x1, ones, N_SAMPLES):
    N_COMPONENTS = 25
    r0, r1 = x0, x1
    r2 = (r0 + r1) / 2

    R0 = torch.rand(N_SAMPLES, N_COMPONENTS)
    R1 = torch.rand(N_SAMPLES, N_COMPONENTS)
    R2 = torch.rand(N_SAMPLES, N_COMPONENTS)

    condition = DirichletBVPSphericalBasis(r_0=r0, R_0=R0, r_1=r1, R_1=R1)
    net = FCNN(1, N_COMPONENTS)

    r = r0 * ones
    assert all_close(condition.enforce(net, r), R0), "inner Dirichlet BC not satisfied"
    r = r1 * ones
    assert all_close(condition.enforce(net, r), R1), "outer Dirichlet BC not satisfied"

    condition = DirichletBVPSphericalBasis(r_0=r2, R_0=R2)
    r = r2 * ones
    assert all_close(condition.enforce(net, r), R2), "single ended BC not satisfied"


def test_inf_dirichlet_bvp_spherical_basis(ones, N_SAMPLES):
    N_COMPONENTS = 25
    r0 = random.random()
    r_inf = 1e15

    R0 = torch.rand(N_SAMPLES, N_COMPONENTS)
    R_inf = torch.rand(N_SAMPLES, N_COMPONENTS)

    condition = InfDirichletBVPSphericalBasis(r_0=r0, R_0=R0, R_inf=R_inf)
    net = FCNN(1, N_COMPONENTS)

    r = r0 * ones
    assert all_close(condition.enforce(net, r), R0), "inner Dirichlet BC not satisfied"
    r = r_inf * ones
    assert all_close(condition.enforce(net, r), R_inf), "Infinity Dirichlet BC not satisfied"


def test_double_ended_bvp_1d(x0, x1, u00, u01, u10, u11, ones, net11):
    u0, u0_prime = u00, u01
    u1, u1_prime = u10, u11
    # test Dirichlet-Dirichlet
    condition = DoubleEndedBVP1D(x_min=x0, x_max=x1, x_min_val=u0, x_max_val=u1)
    x = x0 * ones
    assert all_close(condition.enforce(net11, x), u0), 'left Dirichlet BC not satisfied'
    x = x1 * ones
    assert all_close(condition.enforce(net11, x), u1), 'right Dirichlet BC not satisfied'
    # test Dirichlet-Neumann
    condition = DoubleEndedBVP1D(x_min=x0, x_max=x1, x_min_val=u0, x_max_prime=u1_prime)
    x = x0 * ones
    assert all_close(condition.enforce(net11, x), u0), 'left Dirichlet BC not satisfied'
    x = x1 * ones
    assert all_close(diff(condition.enforce(net11, x), x), u1_prime), 'right Neumann BC not satisfied'
    # test Neumann-Dirichlet
    condition = DoubleEndedBVP1D(x_min=x0, x_max=x1, x_min_prime=u0_prime, x_max_val=u1)
    x = x0 * ones
    assert all_close(diff(condition.enforce(net11, x), x), u0_prime), 'left Neumann BC not satisfied'
    x = x1 * ones
    assert all_close(condition.enforce(net11, x), u1), 'right Dirichlet BC not satisfied'
    # test Neumann-Neumann
    condition = DoubleEndedBVP1D(x_min=x0, x_max=x1, x_min_prime=u0_prime, x_max_prime=u1_prime)
    x = x0 * ones
    assert all_close(diff(condition.enforce(net11, x), x), u0_prime), 'left Neumann BC not satisfied'
    x = x1 * ones
    assert all_close(diff(condition.enforce(net11, x), x), u1_prime), 'right Neumann BC not satisfied'



from random import random
import torch
from torch.utils.tensorboard import SummaryWriter
import dill
import shutil
from pathlib import Path
import os
import pytest
from neurodiffeq import diff
from neurodiffeq.conditions import NoCondition
from neurodiffeq.solvers import Solver1D
from neurodiffeq.monitors import Monitor1D
from neurodiffeq.callbacks import MonitorCallback, CheckpointCallback, ReportOnFitCallback, \
    BaseCallback, SimpleTensorboardCallback, ReportCallback, ActionCallback
from neurodiffeq.callbacks import TrueCallback, FalseCallback, ConditionCallback
from neurodiffeq.callbacks import OnFirstLocal, OnFirstGlobal, OnLastLocal, PeriodGlobal, PeriodLocal
from neurodiffeq.callbacks import ClosedIntervalGlobal, ClosedIntervalLocal, Random
from neurodiffeq.callbacks import RepeatedMetricDown, RepeatedMetricUp, RepeatedMetricDiverge, RepeatedMetricConverge
from neurodiffeq.callbacks import _RepeatedMetricChange
from neurodiffeq.callbacks import EveCallback, StopCallback


@pytest.fixture
def tmp_dir():
    tmp = Path('.') / 'test-tmp'
    yield tmp
    if tmp.exists() and tmp.is_dir():
        shutil.rmtree(tmp)


@pytest.fixture
def solver():
    return Solver1D(
        ode_system=lambda u, t: [u + diff(u, t)],
        conditions=[NoCondition()],
        t_min=0.0,
        t_max=1.0,
    )


@pytest.fixture
def dummy_cb():
    class PassCallback(ActionCallback):
        def __call__(self, solver):
            pass

    return PassCallback()


@pytest.fixture
def true_cb():
    return TrueCallback()


@pytest.fixture
def false_cb():
    return FalseCallback()


def _set_global_epoch(solver, epoch):
    solver.metrics_history['train_loss'] = [0.0] * epoch


def test_monitor_callback(solver, tmp_dir):
    monitor = Monitor1D(0, 1, check_every=30)
    with pytest.warns(FutureWarning):
        MonitorCallback(monitor, check_against_local=True)
    with pytest.warns(FutureWarning):
        MonitorCallback(monitor, check_against='local')
    with pytest.warns(FutureWarning):
        MonitorCallback(monitor, check_against='global')
    with pytest.warns(FutureWarning):
        MonitorCallback(monitor, repaint_last=True)

    callback = MonitorCallback(monitor, fig_dir=tmp_dir)
    assert callback.logger
    callback(solver)

    assert tmp_dir.exists() and tmp_dir.is_dir()


def test_checkpoint_callback(solver, tmp_dir):
    callback = CheckpointCallback(ckpt_dir=tmp_dir)
    callback(solver)
    content = os.listdir(tmp_dir)
    assert len(content) == 1 and content[0].endswith('.internals')

    with open(tmp_dir / content[0], 'rb') as f:
        internals = dill.load(f)

    assert isinstance(internals, dict)
    assert isinstance(internals.get('nets'), list)
    for net in internals.get('nets'):
        assert isinstance(net, torch.nn.Module)


def test_report_callback(solver):
    callback = ReportCallback()
    callback(solver)
    with pytest.warns(FutureWarning):
        callback = ReportOnFitCallback()
        callback(solver)


def test_true_callback(solver, true_cb):
    assert true_cb.condition(solver)


def test_false_callback(solver, false_cb):
    assert not false_cb.condition(dummy_cb)


def test_and_callback(solver, true_cb, false_cb):
    assert (true_cb & true_cb).condition(solver)
    assert not (false_cb & false_cb).condition(solver)
    assert not (true_cb & false_cb).condition(solver)
    assert not (false_cb & true_cb).condition(solver)


def test_or_callback(solver, true_cb, false_cb):
    assert (true_cb | true_cb).condition(solver)
    assert (true_cb | false_cb).condition(solver)
    assert (false_cb | true_cb).condition(solver)
    assert not (false_cb | false_cb).condition(solver)


def test_not_callback(solver, true_cb, false_cb):
    assert not (~true_cb).condition(solver)
    assert (~false_cb).condition(solver)


def test_xor_callback(solver, true_cb, false_cb):
    assert not (true_cb ^ true_cb).condition(solver)
    assert (true_cb ^ false_cb).condition(solver)
    assert (false_cb ^ true_cb).condition(solver)
    assert not (false_cb ^ false_cb).condition(solver)


def test_on_first_local(solver):
    solver.local_epoch = 1
    assert OnFirstLocal().condition(solver)
    solver.local_epoch = 2
    assert not OnFirstLocal().condition(solver)


def test_on_first_global(solver):
    _set_global_epoch(solver, 1)
    assert OnFirstGlobal().condition(solver)
    _set_global_epoch(solver, 2)
    assert not OnFirstGlobal().condition(solver)


def test_on_last_local(solver):
    solver.local_epoch = 9
    solver._max_local_epoch = 10
    assert not OnLastLocal().condition(solver)
    solver.local_epoch = 10
    assert OnLastLocal().condition(solver)


def test_period_local(solver):
    period = 3
    n_periods = 10
    for i in range(1, n_periods):
        solver.local_epoch = i * period
        assert PeriodLocal(period=period).condition(solver)

    for offset in range(1, period):
        for i in range(n_periods):
            solver.local_epoch = i * period + offset
            assert not PeriodLocal(period=period).condition(solver)


def test_period_global(solver):
    period = 3
    n_periods = 10
    for i in range(1, n_periods):
        _set_global_epoch(solver, i * period)
        assert PeriodGlobal(period=period).condition(solver)

    for offset in range(1, period):
        for i in range(n_periods):
            _set_global_epoch(solver, i * period + offset)
            assert not PeriodGlobal(period=period).condition(solver)


def test_closed_interval_local(solver):
    e_min, e_max = [5, 7]
    test_range = list(range(1, 10))
    callback = ClosedIntervalLocal(min=e_min, max=e_max)
    for e in test_range:
        solver.local_epoch = e
        assert callback.condition(solver) == (e_min <= e <= e_max)

    callback = ClosedIntervalLocal(min=e_min)
    for e in test_range:
        solver.local_epoch = e
        assert callback.condition(solver) == (e_min <= e)

    callback = ClosedIntervalLocal(max=e_max)
    for e in test_range:
        solver.local_epoch = e
        assert callback.condition(solver) == (e <= e_max)


def test_closed_interval_global(solver):
    e_min, e_max = [5, 7]
    test_range = list(range(1, 10))
    callback = ClosedIntervalGlobal(min=e_min, max=e_max)
    for e in test_range:
        _set_global_epoch(solver, e)
        assert callback.condition(solver) == (e_min <= e <= e_max)

    callback = ClosedIntervalGlobal(min=e_min)
    for e in test_range:
        _set_global_epoch(solver, e)
        assert callback.condition(solver) == (e_min <= e)

    callback = ClosedIntervalGlobal(max=e_max)
    for e in test_range:
        _set_global_epoch(solver, e)
        assert callback.condition(solver) == (e <= e_max)


def test_random(solver):
    callback = Random(0.5)
    callback.condition(solver)
    Random(0.0)
    Random(1.0)
    with pytest.raises(ValueError):
        Random(-0.1)
    with pytest.raises(ValueError):
        Random(1.1)


def test_repeated_metric_change(solver):
    class Callback(_RepeatedMetricChange):
        def _last_satisfied(self, last, second2last):
            return True

    for phase in ['train', 'valid']:
        use_train = (phase == 'train')
        for metric in ['loss', 'mse', 'something_else']:
            for repetition in range(1, 5):
                callback = Callback(use_train=use_train, metric=metric, repetition=repetition)
                assert callback.key == f'{phase}_{metric}'
                assert callback.times_required == repetition
                assert callback.so_far == 0


def test_repeated_metric_down(solver):
    repetition = 10
    value = 0.0
    at_least_by = 1.0
    callback = RepeatedMetricDown(at_least_by=at_least_by, use_train=True, metric='loss', repetition=repetition)
    for i in range(repetition + 1):
        value -= at_least_by
        solver.metrics_history['train_loss'].append(value)
        if i == repetition:
            assert callback.condition(solver)
        else:
            assert not callback.condition(solver)


def test_repeated_metric_up(solver):
    repetition = 10
    value = 0.0
    at_least_by = 1.0
    callback = RepeatedMetricUp(at_least_by=at_least_by, use_train=True, metric='loss', repetition=repetition)
    for i in range(repetition + 1):
        value += at_least_by
        solver.metrics_history['train_loss'].append(value)
        if i == repetition:
            assert callback.condition(solver)
        else:
            assert not callback.condition(solver)


def test_repeated_converge(solver):
    repetition = 10
    value = 0.0
    epsilon = 0.01
    callback = RepeatedMetricConverge(epsilon=epsilon, use_train=True, metric='loss', repetition=repetition)
    for i in range(repetition + 1):
        value += random() * epsilon * (-1) ** int(random() < 0.5)
        solver.metrics_history['train_loss'].append(value)
        if i == repetition:
            assert callback.condition(solver)
        else:
            assert not callback.condition(solver)


def test_repeated_diverge(solver):
    repetition = 10
    value = 0.0
    gap = 10.0
    callback = RepeatedMetricDiverge(gap=gap, use_train=True, metric='loss', repetition=repetition)
    for i in range(repetition + 1):
        value += (gap / random()) * (-1) ** int(random() < 0.5)
        solver.metrics_history['train_loss'].append(value)
        if i == repetition:
            assert callback.condition(solver)
        else:
            assert not callback.condition(solver)


def test_eve_callback(solver):
    BASE_VALUE = 1000.0
    DOUBLE_AT = 0.5
    N_0 = 3
    callback = EveCallback(base_value=BASE_VALUE, double_at=DOUBLE_AT, n_0=N_0)
    for i in range(5):
        solver.metrics_history['train_loss'] = [BASE_VALUE * (DOUBLE_AT ** i)]
        callback(solver)
        assert solver.n_batches['train'] == (2 ** i) * N_0

    N_MAX = 16
    callback = EveCallback(base_value=BASE_VALUE, double_at=DOUBLE_AT, n_max=N_MAX)
    solver.metrics_history['train_loss'] = [BASE_VALUE * (DOUBLE_AT ** 10)]
    callback(solver)
    assert solver.n_batches['train'] == N_MAX


def test_stop_callback(solver):
    callback = StopCallback()
    callback(solver)
    assert solver._stop_training


def test_tensorboard_callback(solver, tmp_dir):
    writer = SummaryWriter(tmp_dir)
    callback = SimpleTensorboardCallback(writer=writer)
    for i in range(10):
        solver.metrics_history['train_loss'].append(float(i))
        callback(solver)

    callback = SimpleTensorboardCallback()
    for i in range(10):
        solver.metrics_history['train_loss'].append(float(i))
        callback(solver)

    default_path = Path('.') / 'runs'
    assert os.path.isdir(default_path)
    shutil.rmtree(default_path, ignore_errors=True)



from pathlib import Path
import torch


def set_tensor_type(device=None, float_bits=32):
    """Set the default torch tensor type to be used with neurodiffeq.

    :param device: Either "cpu" or "cuda" ("gpu"); defaults to "cuda" if available.
    :type device: str
    :param float_bits: Length of float numbers. Either 32 (float) or 64 (double); defaults to 32.
    :type float_bits: int

    .. note:
        The function calls ``torch.set_default_tensor_type`` under the hood.
        Therefore the ``device`` and ``float_bits`` also becomes default tensor type for PyTorch.
    """
    if not isinstance(float_bits, int):
        raise ValueError(f"float_bits must be int, got {type(float_bits)}")
    if float_bits == 32:
        tensor_str = "FloatTensor"
    elif float_bits == 64:
        tensor_str = "DoubleTensor"
    else:
        raise ValueError(f"float_bits must be 32 or 64, got {float_bits}")

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    if device == "cpu":
        type_string = f"torch.{tensor_str}"
    elif device in ["cuda", "gpu"]:
        type_string = f"torch.cuda.{tensor_str}"
    else:
        raise ValueError(f"Unknown device '{device}'; device must be either 'cuda' or 'cpu'")

    torch.set_default_tensor_type(type_string)


def safe_mkdir(path):
    Path(path).mkdir(parents=True, exist_ok=True)


def get_residual_info(solution, data, diff_eqs, highest_order=0, detach=True):
    # XXX this function is not tested
    from .neurodiffeq import diff

    funcs = solution(data)
    residuals = diff_eqs(*funcs, *data)

    ret = [residuals]
    for order in range(1, highest_order + 1):
        ret.append([[diff(pdr, x) for pdr, x in zip(prev_drs, data)] for prev_drs in ret[-1]])

    if detach:
        def recurse(l):
            for i, entry in enumerate(l):
                if isinstance(entry, torch.Tensor):
                    l[i] = entry.detach()
                else:
                    recurse(entry)

        recurse(ret)

    return ret
